from cryptography.hazmat.primitives.hashes import HashAlgorithm
from cryptography.x509 import ObjectIdentifier

class ExtensionOID:
    SUBJECT_DIRECTORY_ATTRIBUTES: ObjectIdentifier = ...
    SUBJECT_KEY_IDENTIFIER: ObjectIdentifier = ...
    KEY_USAGE: ObjectIdentifier = ...
    SUBJECT_ALTERNATIVE_NAME: ObjectIdentifier = ...
    ISSUER_ALTERNATIVE_NAME: ObjectIdentifier = ...
    BASIC_CONSTRAINTS: ObjectIdentifier = ...
    NAME_CONSTRAINTS: ObjectIdentifier = ...
    CRL_DISTRIBUTION_POINTS: ObjectIdentifier = ...
    CERTIFICATE_POLICIES: ObjectIdentifier = ...
    POLICY_MAPPINGS: ObjectIdentifier = ...
    AUTHORITY_KEY_IDENTIFIER: ObjectIdentifier = ...
    POLICY_CONSTRAINTS: ObjectIdentifier = ...
    EXTENDED_KEY_USAGE: ObjectIdentifier = ...
    FRESHEST_CRL: ObjectIdentifier = ...
    INHIBIT_ANY_POLICY: ObjectIdentifier = ...
    ISSUING_DISTRIBUTION_POINT: ObjectIdentifier = ...
    AUTHORITY_INFORMATION_ACCESS: ObjectIdentifier = ...
    SUBJECT_INFORMATION_ACCESS: ObjectIdentifier = ...
    OCSP_NO_CHECK: ObjectIdentifier = ...
    TLS_FEATURE: ObjectIdentifier = ...
    CRL_NUMBER: ObjectIdentifier = ...
    DELTA_CRL_INDICATOR: ObjectIdentifier = ...
    PRECERT_SIGNED_CERTIFICATE_TIMESTAMPS: ObjectIdentifier = ...
    PRECERT_POISON: ObjectIdentifier = ...

class OCSPExtensionOID:
    NONCE: ObjectIdentifier = ...

class CRLEntryExtensionOID:
    CERTIFICATE_ISSUER: ObjectIdentifier = ...
    CRL_REASON: ObjectIdentifier = ...
    INVALIDITY_DATE: ObjectIdentifier = ...

class NameOID:
    COMMON_NAME: ObjectIdentifier = ...
    COUNTRY_NAME: ObjectIdentifier = ...
    LOCALITY_NAME: ObjectIdentifier = ...
    STATE_OR_PROVINCE_NAME: ObjectIdentifier = ...
    STREET_ADDRESS: ObjectIdentifier = ...
    ORGANIZATION_NAME: ObjectIdentifier = ...
    ORGANIZATIONAL_UNIT_NAME: ObjectIdentifier = ...
    SERIAL_NUMBER: ObjectIdentifier = ...
    SURNAME: ObjectIdentifier = ...
    GIVEN_NAME: ObjectIdentifier = ...
    TITLE: ObjectIdentifier = ...
    GENERATION_QUALIFIER: ObjectIdentifier = ...
    X500_UNIQUE_IDENTIFIER: ObjectIdentifier = ...
    DN_QUALIFIER: ObjectIdentifier = ...
    PSEUDONYM: ObjectIdentifier = ...
    USER_ID: ObjectIdentifier = ...
    DOMAIN_COMPONENT: ObjectIdentifier = ...
    EMAIL_ADDRESS: ObjectIdentifier = ...
    JURISDICTION_COUNTRY_NAME: ObjectIdentifier = ...
    JURISDICTION_LOCALITY_NAME: ObjectIdentifier = ...
    JURISDICTION_STATE_OR_PROVINCE_NAME: ObjectIdentifier = ...
    BUSINESS_CATEGORY: ObjectIdentifier = ...
    POSTAL_ADDRESS: ObjectIdentifier = ...
    POSTAL_CODE: ObjectIdentifier = ...

class SignatureAlgorithmOID:
    RSA_WITH_MD5: ObjectIdentifier = ...
    RSA_WITH_SHA1: ObjectIdentifier = ...
    _RSA_WITH_SHA1: ObjectIdentifier = ...
    RSA_WITH_SHA224: ObjectIdentifier = ...
    RSA_WITH_SHA256: ObjectIdentifier = ...
    RSA_WITH_SHA384: ObjectIdentifier = ...
    RSA_WITH_SHA512: ObjectIdentifier = ...
    RSASSA_PSS: ObjectIdentifier = ...
    ECDSA_WITH_SHA1: ObjectIdentifier = ...
    ECDSA_WITH_SHA224: ObjectIdentifier = ...
    ECDSA_WITH_SHA256: ObjectIdentifier = ...
    ECDSA_WITH_SHA384: ObjectIdentifier = ...
    ECDSA_WITH_SHA512: ObjectIdentifier = ...
    DSA_WITH_SHA1: ObjectIdentifier = ...
    DSA_WITH_SHA224: ObjectIdentifier = ...
    DSA_WITH_SHA256: ObjectIdentifier = ...
    ED25519: ObjectIdentifier = ...
    ED448: ObjectIdentifier = ...

class ExtendedKeyUsageOID:
    SERVER_AUTH: ObjectIdentifier = ...
    CLIENT_AUTH: ObjectIdentifier = ...
    CODE_SIGNING: ObjectIdentifier = ...
    EMAIL_PROTECTION: ObjectIdentifier = ...
    TIME_STAMPING: ObjectIdentifier = ...
    OCSP_SIGNING: ObjectIdentifier = ...
    ANY_EXTENDED_KEY_USAGE: ObjectIdentifier = ...

class AuthorityInformationAccessOID:
    CA_ISSUERS: ObjectIdentifier = ...
    OCSP: ObjectIdentifier = ...

class CertificatePoliciesOID:
    CPS_QUALIFIER: ObjectIdentifier = ...
    CPS_USER_NOTICE: ObjectIdentifier = ...
    ANY_POLICY: ObjectIdentifier = ...

_OID_NAMES: dict[ObjectIdentifier, str] = ...

_SIG_OIDS_TO_HASH: dict[ObjectIdentifier, HashAlgorithm | None] = ...



"""File and path utilities."""

import contextlib
import errno
import glob
import os
import shutil
import tempfile
import textwrap


def recursive_glob(path):
  """Call recursive glob iff ** is in the pattern."""
  if "*" not in path:
    # Glob isn't needed.
    return [path]
  elif "**" not in path:
    # Recursive glob isn't needed.
    return glob.glob(path)
  else:
    return glob.glob(path, recursive=True)


def replace_extension(filename, new_extension):
  name, _ = os.path.splitext(filename)
  if new_extension.startswith("."):
    return name + new_extension
  else:
    return name + "." + new_extension


def makedirs(path):
  """Create a nested directory, but don't fail if any of it already exists."""
  try:
    os.makedirs(path)
  except OSError as e:
    if e.errno != errno.EEXIST:
      raise


class Tempdir:
  """Context handler for creating temporary directories."""

  def __enter__(self):
    self.path = tempfile.mkdtemp()
    return self

  def create_directory(self, filename):
    """Create a subdirectory in the temporary directory."""
    path = os.path.join(self.path, filename)
    makedirs(path)
    return path

  def create_file(self, filename, indented_data=None):
    """Create a file in the temporary directory. Dedents the data if needed."""
    filedir, filename = os.path.split(filename)
    if filedir:
      self.create_directory(filedir)
    path = os.path.join(self.path, filedir, filename)
    if isinstance(indented_data, bytes):
      # This is binary data rather than text.
      mode = "wb"
      data = indented_data
    else:
      mode = "w"
      data = textwrap.dedent(indented_data) if indented_data else indented_data
    with open(path, mode) as fi:
      if data:
        fi.write(data)
    return path

  def delete_file(self, filename):
    os.unlink(os.path.join(self.path, filename))

  def __exit__(self, error_type, value, tb):
    shutil.rmtree(path=self.path)
    return False  # reraise any exceptions

  def __getitem__(self, filename):
    """Get the full path for an entry in this directory."""
    return os.path.join(self.path, filename)


@contextlib.contextmanager
def cd(path):
  """Context manager. Change the directory, and restore it afterwards.

  Example usage:
    with cd("/path"):
      ...

  Arguments:
    path: The directory to change to. If empty, this function is a no-op.
  Yields:
    Executes your code, in a changed directory.
  """
  if not path:
    yield
    return
  curdir = os.getcwd()
  os.chdir(path)
  try:
    yield
  finally:
    os.chdir(curdir)


def is_pyi_directory_init(filename):
  """Checks if a pyi file is path/to/dir/__init__.pyi."""
  if filename is None:
    return False
  return os.path.splitext(os.path.basename(filename))[0] == "__init__"


def expand_path(path, cwd=None):
  """Fully expand a path, optionally with an explicit cwd."""

  expand = lambda path: os.path.realpath(os.path.expanduser(path))
  with cd(cwd):
    return expand(path)


def expand_paths(paths, cwd=None):
  """Fully expand a list of paths, optionally with an explicit cwd."""
  return [expand_path(x, cwd) for x in paths]


def expand_globpaths(globpaths, cwd=None):
  """Expand a list of glob expressions into a list of full paths."""
  with cd(cwd):
    paths = sum((recursive_glob(p) for p in globpaths), [])
  return expand_paths(paths, cwd)


def expand_source_files(filenames, cwd=None):
  """Expand a space-separated string of filenames passed in as sources.

  This is a helper function for handling command line arguments that specify a
  list of source files and directories.

  Any directories in filenames will be scanned recursively for .py files.
  Any files that do not end with ".py" will be dropped.

  Args:
    filenames: A space-separated string of filenames to process.
    cwd: An optional working directory to expand relative paths
  Returns:
    A set of full paths to .py files
  """
  out = []
  for f in expand_globpaths(filenames.split(), cwd):
    if os.path.isdir(f):
      # If we have a directory, collect all the .py files within it.
      out += recursive_glob(os.path.join(f, "**", "*.py"))
    elif f.endswith(".py"):
      out.append(f)
  return set(out)


def expand_pythonpath(pythonpath, cwd=None):
  """Expand a/b:c/d into [/path/to/a/b, /path/to/c/d]."""
  if pythonpath:
    return expand_paths(
        (path.strip() for path in pythonpath.split(os.pathsep)), cwd)
  else:
    return []



"""Tests for convert.py."""

from pytype import config
from pytype import context
from pytype import errors
from pytype import file_utils
from pytype import load_pytd
from pytype.abstract import abstract
from pytype.abstract import abstract_utils
from pytype.pytd import pytd
from pytype.pytd import pytd_utils
from pytype.tests import test_base

import unittest


class ConvertTest(test_base.UnitTest):

  def setUp(self):
    super().setUp()
    options = config.Options.create(python_version=self.python_version)
    self._ctx = context.Context(errors.ErrorLog(), options,
                                load_pytd.Loader(options))

  def _load_ast(self, name, src):
    with file_utils.Tempdir() as d:
      d.create_file(name + ".pyi", src)
      self._ctx.options.tweak(pythonpath=[d.path])  # monkeypatch
      return self._ctx.loader.import_name(name)

  def _convert_class(self, name, ast):
    return self._ctx.convert.constant_to_value(
        ast.Lookup(name), {}, self._ctx.root_node)

  def test_convert_metaclass(self):
    ast = self._load_ast("a", """
      class A(type): ...
      class B(metaclass=A): ...
      class C(B): ...
    """)
    meta = self._convert_class("a.A", ast)
    cls_meta = self._convert_class("a.B", ast).cls
    subcls_meta = self._convert_class("a.C", ast).cls
    self.assertEqual(meta, cls_meta)
    self.assertEqual(meta, subcls_meta)

  def test_convert_default_metaclass(self):
    ast = self._load_ast("a", """
      class A: ...
    """)
    cls = self._convert_class("a.A", ast)
    self.assertEqual(cls.cls, self._ctx.convert.type_type)

  def test_convert_metaclass_with_generic(self):
    ast = self._load_ast("a", """
      from typing import Generic, TypeVar
      T = TypeVar("T")
      class A(type): ...
      class B(Generic[T], metaclass=A): ...
      class C(B[int]): ...
    """)
    meta = self._convert_class("a.A", ast)
    cls_meta = self._convert_class("a.B", ast).cls
    subcls_meta = self._convert_class("a.C", ast).cls
    self.assertEqual(meta, cls_meta)
    self.assertEqual(meta, subcls_meta)

  def test_generic_with_any_param(self):
    ast = self._load_ast("a", """
      from typing import Dict
      x = ...  # type: Dict[str]
    """)
    val = self._ctx.convert.constant_to_value(
        ast.Lookup("a.x").type, {}, self._ctx.root_node)
    self.assertIs(val.formal_type_parameters[abstract_utils.K],
                  self._ctx.convert.str_type)
    self.assertIs(val.formal_type_parameters[abstract_utils.V],
                  self._ctx.convert.unsolvable)

  def test_convert_long(self):
    val = self._ctx.convert.constant_to_value(2**64, {}, self._ctx.root_node)
    self.assertIs(val, self._ctx.convert.primitive_class_instances[int])

  def test_heterogeneous_tuple(self):
    ast = self._load_ast("a", """
      from typing import Tuple
      x = ...  # type: Tuple[str, int]
    """)
    x = ast.Lookup("a.x").type
    cls = self._ctx.convert.constant_to_value(x, {}, self._ctx.root_node)
    instance = self._ctx.convert.constant_to_value(
        abstract_utils.AsInstance(x), {}, self._ctx.root_node)
    self.assertIsInstance(cls, abstract.TupleClass)
    self.assertCountEqual(cls.formal_type_parameters.items(),
                          [(0, self._ctx.convert.str_type),
                           (1, self._ctx.convert.int_type),
                           (abstract_utils.T,
                            abstract.Union([
                                cls.formal_type_parameters[0],
                                cls.formal_type_parameters[1],
                            ], self._ctx))])
    self.assertIsInstance(instance, abstract.Tuple)
    self.assertListEqual([v.data for v in instance.pyval],
                         [[self._ctx.convert.primitive_class_instances[str]],
                          [self._ctx.convert.primitive_class_instances[int]]])
    # The order of option elements in Union is random
    self.assertCountEqual(
        instance.get_instance_type_parameter(abstract_utils.T).data, [
            self._ctx.convert.primitive_class_instances[str],
            self._ctx.convert.primitive_class_instances[int]
        ])

  def test_build_bool(self):
    any_bool = self._ctx.convert.build_bool(self._ctx.root_node, None)
    t_bool = self._ctx.convert.build_bool(self._ctx.root_node, True)
    f_bool = self._ctx.convert.build_bool(self._ctx.root_node, False)
    self.assertEqual(any_bool.data,
                     [self._ctx.convert.primitive_class_instances[bool]])
    self.assertEqual(t_bool.data, [self._ctx.convert.true])
    self.assertEqual(f_bool.data, [self._ctx.convert.false])

  def test_boolean_constants(self):
    true = self._ctx.convert.constant_to_value(True, {}, self._ctx.root_node)
    self.assertEqual(true, self._ctx.convert.true)
    false = self._ctx.convert.constant_to_value(False, {}, self._ctx.root_node)
    self.assertEqual(false, self._ctx.convert.false)

  def test_callable_with_args(self):
    ast = self._load_ast("a", """
      from typing import Callable
      x = ...  # type: Callable[[int, bool], str]
    """)
    x = ast.Lookup("a.x").type
    cls = self._ctx.convert.constant_to_value(x, {}, self._ctx.root_node)
    instance = self._ctx.convert.constant_to_value(
        abstract_utils.AsInstance(x), {}, self._ctx.root_node)
    self.assertIsInstance(cls, abstract.CallableClass)
    self.assertCountEqual(
        cls.formal_type_parameters.items(),
        [(0, self._ctx.convert.int_type),
         (1, self._ctx.convert.primitive_classes[bool]),
         (abstract_utils.ARGS,
          abstract.Union(
              [cls.formal_type_parameters[0], cls.formal_type_parameters[1]],
              self._ctx)), (abstract_utils.RET, self._ctx.convert.str_type)])
    self.assertIsInstance(instance, abstract.Instance)
    self.assertEqual(instance.cls, cls)
    self.assertCountEqual(
        [(name, set(var.data))
         for name, var in instance.instance_type_parameters.items()],
        [(abstract_utils.full_type_name(instance, abstract_utils.ARGS), {
            self._ctx.convert.primitive_class_instances[int],
            self._ctx.convert.primitive_class_instances[bool]
        }),
         (abstract_utils.full_type_name(instance, abstract_utils.RET),
          {self._ctx.convert.primitive_class_instances[str]})])

  def test_callable_no_args(self):
    ast = self._load_ast("a", """
      from typing import Callable
      x = ... # type: Callable[[], ...]
    """)
    x = ast.Lookup("a.x").type
    cls = self._ctx.convert.constant_to_value(x, {}, self._ctx.root_node)
    instance = self._ctx.convert.constant_to_value(
        abstract_utils.AsInstance(x), {}, self._ctx.root_node)
    self.assertIsInstance(
        cls.get_formal_type_parameter(abstract_utils.ARGS), abstract.Empty)
    self.assertEqual(
        abstract_utils.get_atomic_value(
            instance.get_instance_type_parameter(abstract_utils.ARGS)),
        self._ctx.convert.empty)

  def test_plain_callable(self):
    ast = self._load_ast("a", """
      from typing import Callable
      x = ...  # type: Callable[..., int]
    """)
    x = ast.Lookup("a.x").type
    cls = self._ctx.convert.constant_to_value(x, {}, self._ctx.root_node)
    instance = self._ctx.convert.constant_to_value(
        abstract_utils.AsInstance(x), {}, self._ctx.root_node)
    self.assertIsInstance(cls, abstract.ParameterizedClass)
    self.assertCountEqual(cls.formal_type_parameters.items(),
                          [(abstract_utils.ARGS, self._ctx.convert.unsolvable),
                           (abstract_utils.RET, self._ctx.convert.int_type)])
    self.assertIsInstance(instance, abstract.Instance)
    self.assertEqual(instance.cls, cls.base_cls)
    self.assertCountEqual(
        [(name, var.data)
         for name, var in instance.instance_type_parameters.items()],
        [(abstract_utils.full_type_name(
            instance, abstract_utils.ARGS), [self._ctx.convert.unsolvable]),
         (abstract_utils.full_type_name(instance, abstract_utils.RET),
          [self._ctx.convert.primitive_class_instances[int]])])

  def test_function_with_starargs(self):
    ast = self._load_ast("a", """
      def f(*args: int): ...
    """)
    f = self._ctx.convert.constant_to_value(
        ast.Lookup("a.f"), {}, self._ctx.root_node)
    sig, = f.signatures
    annot = sig.signature.annotations["args"]
    self.assertEqual(pytd_utils.Print(annot.get_instance_type()),
                     "Tuple[int, ...]")

  def test_function_with_starstarargs(self):
    ast = self._load_ast("a", """
      def f(**kwargs: int): ...
    """)
    f = self._ctx.convert.constant_to_value(
        ast.Lookup("a.f"), {}, self._ctx.root_node)
    sig, = f.signatures
    annot = sig.signature.annotations["kwargs"]
    self.assertEqual(pytd_utils.Print(annot.get_instance_type()),
                     "Dict[str, int]")

  def test_mro(self):
    ast = self._load_ast("a", """
      x = ...  # type: dict
    """)
    x = ast.Lookup("a.x").type
    cls = self._ctx.convert.constant_to_value(x, {}, self._ctx.root_node)
    self.assertListEqual([v.name for v in cls.mro],
                         ["dict", "Dict", "MutableMapping", "Mapping", "Sized",
                          "Iterable", "Container", "Generic", "Protocol",
                          "object"])

  def test_widen_type(self):
    ast = self._load_ast("a", """
      x = ...  # type: tuple[int, ...]
      y = ...  # type: dict[str, int]
    """)
    x = ast.Lookup("a.x").type
    tup = self._ctx.convert.constant_to_value(x, {}, self._ctx.root_node)
    widened_tup = self._ctx.convert.widen_type(tup)
    self.assertEqual(pytd_utils.Print(widened_tup.get_instance_type()),
                     "Iterable[int]")
    y = ast.Lookup("a.y").type
    dct = self._ctx.convert.constant_to_value(y, {}, self._ctx.root_node)
    widened_dct = self._ctx.convert.widen_type(dct)
    self.assertEqual(pytd_utils.Print(widened_dct.get_instance_type()),
                     "Mapping[str, int]")

  def test_abstract_method_round_trip(self):
    sig = pytd.Signature((), None, None, pytd.AnythingType(), (), ())
    f_pytd = pytd.Function(
        name="f", signatures=(sig,), kind=pytd.MethodTypes.METHOD,
        flags=pytd.MethodFlags.ABSTRACT)
    f = self._ctx.convert.constant_to_value(f_pytd, {}, self._ctx.root_node)
    self.assertTrue(f.is_abstract)
    f_out = f.to_pytd_def(self._ctx.root_node, f.name)
    self.assertTrue(f_out.is_abstract)

  def test_class_abstract_method(self):
    ast = self._load_ast("a", """
      class A:
        @abstractmethod
        def f(self) -> int: ...
    """)
    cls = self._ctx.convert.constant_to_value(
        ast.Lookup("a.A"), {}, self._ctx.root_node)
    self.assertCountEqual(cls.abstract_methods, {"f"})

  def test_class_inherited_abstract_method(self):
    ast = self._load_ast("a", """
      class A:
        @abstractmethod
        def f(self) -> int: ...
      class B(A): ...
    """)
    cls = self._ctx.convert.constant_to_value(
        ast.Lookup("a.B"), {}, self._ctx.root_node)
    self.assertCountEqual(cls.abstract_methods, {"f"})

  def test_class_override_abstract_method(self):
    ast = self._load_ast("a", """
      class A:
        @abstractmethod
        def f(self) -> int: ...
      class B(A):
        def f(self) -> bool: ...
    """)
    cls = self._ctx.convert.constant_to_value(
        ast.Lookup("a.B"), {}, self._ctx.root_node)
    self.assertFalse(cls.abstract_methods)

  def test_class_override_abstract_method_still_abstract(self):
    ast = self._load_ast("a", """
      class A:
        @abstractmethod
        def f(self) -> int: ...
      class B(A):
        @abstractmethod
        def f(self) -> bool: ...
    """)
    cls = self._ctx.convert.constant_to_value(
        ast.Lookup("a.B"), {}, self._ctx.root_node)
    self.assertCountEqual(cls.abstract_methods, {"f"})

  def test_parameterized_class_abstract_method(self):
    ast = self._load_ast("a", """
      class A:
        @abstractmethod
        def f(self) -> int: ...
    """)
    cls = self._ctx.convert.constant_to_value(
        ast.Lookup("a.A"), {}, self._ctx.root_node)
    parameterized_cls = abstract.ParameterizedClass(cls, {}, self._ctx)
    self.assertCountEqual(parameterized_cls.abstract_methods, {"f"})

  def test_classvar(self):
    ast = self._load_ast("a", """
      from typing import ClassVar
      class X:
        v: ClassVar[int]
    """)
    pyval = ast.Lookup("a.X").Lookup("v").type
    v = self._ctx.convert.constant_to_value(pyval, {}, self._ctx.root_node)
    self.assertEqual(v, self._ctx.convert.int_type)

  def test_classvar_instance(self):
    ast = self._load_ast("a", """
      from typing import ClassVar
      class X:
        v: ClassVar[int]
    """)
    pyval = ast.Lookup("a.X").Lookup("v").type
    v = self._ctx.convert.constant_to_value(
        abstract_utils.AsInstance(pyval), {}, self._ctx.root_node)
    self.assertEqual(v, self._ctx.convert.primitive_class_instances[int])

  def test_constant_name(self):
    # Test that we create a string name without crashing.
    self.assertIsInstance(self._ctx.convert.constant_name(int), str)
    self.assertIsInstance(self._ctx.convert.constant_name(None), str)
    self.assertIsInstance(
        self._ctx.convert.constant_name((int, (str, super))), str)


if __name__ == "__main__":
  unittest.main()



from typing import Text

class InvalidToken(Exception): ...

class Fernet(object):
    def __init__(self, key: bytes | Text) -> None: ...
    def decrypt(self, token: bytes, ttl: int | None = ...) -> bytes: ...
    # decrypt_at_time accepts None ttl at runtime but it's an implementtion detail and it doesn't really
    # make sense for the client code to use it like that, so the parameter is typed as int as opposed to
    # int | None.
    def decrypt_at_time(self, token: bytes, ttl: int, current_time: int) -> bytes: ...
    def encrypt(self, data: bytes) -> bytes: ...
    def encrypt_at_time(self, data: bytes, current_time: int) -> bytes: ...
    def extract_timestamp(self, token: bytes) -> int: ...
    @classmethod
    def generate_key(cls) -> bytes: ...

class MultiFernet(object):
    def __init__(self, fernets: list[Fernet]) -> None: ...
    def decrypt(self, token: bytes, ttl: int | None = ...) -> bytes: ...
    # See a note above on the typing of the ttl parameter.
    def decrypt_at_time(self, token: bytes, ttl: int, current_time: int) -> bytes: ...
    def encrypt(self, data: bytes) -> bytes: ...
    def encrypt_at_time(self, data: bytes, current_time: int) -> bytes: ...
    def rotate(self, msg: bytes) -> bytes: ...



from logging import Logger
from typing import Any

from aws_xray_sdk import global_sdk_config as global_sdk_config

from .context import Context as Context
from .models.facade_segment import FacadeSegment as FacadeSegment
from .models.trace_header import TraceHeader as TraceHeader

log: Logger
LAMBDA_TRACE_HEADER_KEY: str
LAMBDA_TASK_ROOT_KEY: str
TOUCH_FILE_DIR: str
TOUCH_FILE_PATH: str

def check_in_lambda(): ...

class LambdaContext(Context):
    def __init__(self) -> None: ...
    def put_segment(self, segment) -> None: ...
    def end_segment(self, end_time: Any | None = ...) -> None: ...
    def put_subsegment(self, subsegment) -> None: ...
    def get_trace_entity(self): ...
    @property
    def context_missing(self) -> None: ...
    @context_missing.setter
    def context_missing(self, value) -> None: ...
    def handle_context_missing(self) -> None: ...



import sys
from _typeshed import Self
from typing import Any, Iterator, Sequence, Text, Type

if sys.version_info >= (3, 0):
    from configparser import ConfigParser
else:
    from ConfigParser import ConfigParser

if sys.version_info >= (3, 8):
    from re import Pattern
else:
    from typing import Pattern

entry_point_pattern: Pattern[Text]
file_in_zip_pattern: Pattern[Text]

class BadEntryPoint(Exception):
    epstr: Text
    def __init__(self, epstr: Text) -> None: ...
    @staticmethod
    def err_to_warnings() -> Iterator[None]: ...

class NoSuchEntryPoint(Exception):
    group: Text
    name: Text
    def __init__(self, group: Text, name: Text) -> None: ...

class EntryPoint:
    name: Text
    module_name: Text
    object_name: Text
    extras: Sequence[Text] | None
    distro: Distribution | None
    def __init__(
        self,
        name: Text,
        module_name: Text,
        object_name: Text,
        extras: Sequence[Text] | None = ...,
        distro: Distribution | None = ...,
    ) -> None: ...
    def load(self) -> Any: ...
    @classmethod
    def from_string(cls: Type[Self], epstr: Text, name: Text, distro: Distribution | None = ...) -> Self: ...

class Distribution:
    name: Text
    version: Text
    def __init__(self, name: Text, version: Text) -> None: ...

def iter_files_distros(
    path: Sequence[Text] | None = ..., repeated_distro: Text = ...
) -> Iterator[tuple[ConfigParser, Distribution | None]]: ...
def get_single(group: Text, name: Text, path: Sequence[Text] | None = ...) -> EntryPoint: ...
def get_group_named(group: Text, path: Sequence[Text] | None = ...) -> dict[str, EntryPoint]: ...
def get_group_all(group: Text, path: Sequence[Text] | None = ...) -> list[EntryPoint]: ...



"""Test for the cfg Python extension module."""

from pytype.typegraph import cfg

import unittest


class CFGTest(unittest.TestCase):
  """Test control flow graph creation."""

  def test_simple_graph(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("foo")
    n2 = n1.ConnectNew("n2")
    n3 = n1.ConnectNew("n3")
    n4 = n3.ConnectNew("n4")
    self.assertEqual(0, n1.id)
    self.assertEqual("foo", n1.name)
    self.assertEqual(len(n1.outgoing), 2)
    self.assertEqual(len(n2.outgoing), 0)  # pylint: disable=g-generic-assert
    self.assertEqual(len(n3.outgoing), 1)
    self.assertEqual(len(n2.incoming), 1)
    self.assertEqual(len(n3.incoming), 1)
    self.assertEqual(len(n4.incoming), 1)
    self.assertIn(n2, n1.outgoing)
    self.assertIn(n3, n1.outgoing)
    self.assertIn(n1, n2.incoming)
    self.assertIn(n1, n3.incoming)
    self.assertIn(n3, n4.incoming)

  def test_binding_binding(self):
    p = cfg.Program()
    node = p.NewCFGNode()
    u = p.NewVariable()
    v1 = u.AddBinding(None, source_set=[], where=node)
    v2 = u.AddBinding("data", source_set=[], where=node)
    v3 = u.AddBinding({1: 2}, source_set=[], where=node)
    self.assertIsNone(v1.data)
    self.assertEqual(v2.data, "data")
    self.assertEqual(v3.data, {1: 2})
    self.assertEqual(f"<binding of variable 0 to data {id(v3.data)}>", str(v3))

  def test_cfg_node_str(self):
    p = cfg.Program()
    n1 = p.NewCFGNode()
    n2 = p.NewCFGNode("n2")
    v = p.NewVariable()
    av = v.AddBinding("a", source_set=[], where=n1)
    n3 = p.NewCFGNode("n3", condition=av)
    self.assertEqual("<cfgnode 0 None>", str(n1))
    self.assertEqual("<cfgnode 1 n2>", str(n2))
    self.assertEqual("<cfgnode 2 n3 condition:0>", str(n3))

  def test_get_attro(self):
    p = cfg.Program()
    node = p.NewCFGNode()
    u = p.NewVariable()
    data = [1, 2, 3]
    a = u.AddBinding(data, source_set=[], where=node)
    self.assertEqual(a.variable.bindings, [a])
    origin, = a.origins  # we expect exactly one origin
    self.assertEqual(origin.where, node)
    self.assertEqual(len(origin.source_sets), 1)
    source_set, = origin.source_sets
    self.assertEqual(list(source_set), [])
    self.assertEqual(a.data, data)

  def test_get_origins(self):
    p = cfg.Program()
    node = p.NewCFGNode()
    u = p.NewVariable()
    a = u.AddBinding(1, source_set=[], where=node)
    b = u.AddBinding(2, source_set=[a], where=node)
    c = u.AddBinding(3, source_set=[a, b], where=node)
    expected_source_sets = [[], [a], [a, b]]
    for binding, expected_source_set in zip([a, b, c], expected_source_sets):
      origin, = binding.origins
      self.assertEqual(origin.where, node)
      source_set, = origin.source_sets
      self.assertCountEqual(list(source_set), expected_source_set)

  def test_variable_set(self):
    p = cfg.Program()
    node1 = p.NewCFGNode("n1")
    node2 = node1.ConnectNew("n2")
    d = p.NewVariable()
    d.AddBinding("v1", source_set=[], where=node1)
    d.AddBinding("v2", source_set=[], where=node2)
    self.assertEqual(len(d.bindings), 2)

  def test_has_source(self):
    p = cfg.Program()
    n0, n1, n2 = p.NewCFGNode("n0"), p.NewCFGNode("n1"), p.NewCFGNode("n2")
    u = p.NewVariable()
    u1 = u.AddBinding(0, source_set=[], where=n0)
    v = p.NewVariable()
    v1 = v.AddBinding(1, source_set=[], where=n1)
    v2 = v.AddBinding(2, source_set=[u1], where=n1)
    v3a = v.AddBinding(3, source_set=[], where=n1)
    v3b = v.AddBinding(3, source_set=[u1], where=n2)
    self.assertEqual(v3a, v3b)
    v3 = v3a
    self.assertTrue(v1.HasSource(v1))
    self.assertTrue(v2.HasSource(v2))
    self.assertTrue(v3.HasSource(v3))
    self.assertFalse(v1.HasSource(u1))
    self.assertTrue(v2.HasSource(u1))
    self.assertTrue(v3.HasSource(u1))

  def test_filter1(self):
    #                    x.ab = A()
    #               ,---+------------.
    #               |   n3           |
    #  x = X()      |    x.ab = B()  |
    #  +------------+---+------------+------------+
    #  n1           n2  n4           n5           n6
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    n3 = n2.ConnectNew("n3")
    n4 = n2.ConnectNew("n4")
    n5 = n3.ConnectNew("n5")
    n4.ConnectTo(n5)
    n6 = n5.ConnectNew("n6")
    n5.ConnectTo(n6)

    all_x = p.NewVariable()
    x = all_x.AddBinding({}, source_set=[], where=n1)
    ab = p.NewVariable()
    x.data["ab"] = ab
    a = ab.AddBinding("A", source_set=[], where=n3)
    b = ab.AddBinding("B", source_set=[], where=n4)

    p.entrypoint = n1
    self.assertFalse(a.IsVisible(n1) or b.IsVisible(n1))
    self.assertFalse(a.IsVisible(n2) or b.IsVisible(n2))
    self.assertTrue(a.IsVisible(n3))
    self.assertTrue(b.IsVisible(n4))
    self.assertEqual(ab.Filter(n1), [])
    self.assertEqual(ab.Filter(n2), [])
    self.assertEqual(ab.FilteredData(n3), ["A"])
    self.assertEqual(ab.FilteredData(n4), ["B"])
    self.assertCountEqual(["A", "B"], ab.FilteredData(n5))
    self.assertCountEqual(["A", "B"], ab.FilteredData(n6))

  def test_can_have_combination(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    n3 = n1.ConnectNew("n3")
    n4 = p.NewCFGNode("n4")
    n2.ConnectTo(n4)
    n3.ConnectTo(n4)
    x = p.NewVariable()
    y = p.NewVariable()
    x1 = x.AddBinding("1", source_set=[], where=n2)
    y2 = y.AddBinding("2", source_set=[], where=n3)
    self.assertTrue(n4.CanHaveCombination([x1, y2]))
    self.assertTrue(n4.CanHaveCombination([x1]))
    self.assertTrue(n4.CanHaveCombination([y2]))
    self.assertTrue(n3.CanHaveCombination([y2]))
    self.assertTrue(n2.CanHaveCombination([x1]))
    self.assertTrue(n1.CanHaveCombination([]))
    self.assertFalse(n1.CanHaveCombination([x1]))
    self.assertFalse(n1.CanHaveCombination([y2]))
    self.assertFalse(n2.CanHaveCombination([x1, y2]))
    self.assertFalse(n3.CanHaveCombination([x1, y2]))

  def test_conflicting_bindings_from_condition(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    n3 = n2.ConnectNew("n3")
    x = p.NewVariable()
    x_a = x.AddBinding("a", source_set=[], where=n1)
    x_b = x.AddBinding("b", source_set=[], where=n1)
    p.entrypoint = n1
    n2.condition = x_a
    self.assertFalse(n3.HasCombination([x_b]))

  def test_condition_order(self):
    p = cfg.Program()
    x, y = p.NewVariable(), p.NewVariable()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    n3 = n2.ConnectNew("n3")
    n4 = n3.ConnectNew("n4")
    n5 = n4.ConnectNew("n5")
    n6 = n5.ConnectNew("n6")
    p.entrypoint = n1
    y_a = y.AddBinding("a", source_set=[], where=n1)
    n3.condition = x.AddBinding("b", source_set=[], where=n2)
    n5.condition = x.AddBinding("c", source_set=[], where=n4)
    self.assertTrue(n6.HasCombination([y_a]))

  def test_contained_if_conflict(self):
    p = cfg.Program()
    x = p.NewVariable()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    n3 = n1.ConnectNew("n3")
    n4 = p.NewCFGNode("n4")
    n2.ConnectTo(n4)
    n3.ConnectTo(n4)
    n5 = n4.ConnectNew("n5")
    p.entrypoint = n1
    x_a = x.AddBinding("a", source_set=[], where=n1)
    n4.condition = x.AddBinding("b", source_set=[], where=n2)
    # This is impossible since we have a condition on the way, enforcing x=b.
    self.assertFalse(n5.HasCombination([x_a]))

  def test_conflicting_conditions_on_path(self):
    # This test case is rather academic - there's no obvious way to construct
    # a Python program that actually creates the CFG below.
    p = cfg.Program()
    x, y, z = p.NewVariable(), p.NewVariable(), p.NewVariable()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    n3 = n1.ConnectNew("n3")
    n4 = p.NewCFGNode("n4")
    n2.ConnectTo(n4)
    n3.ConnectTo(n4)
    n5 = n4.ConnectNew("n5")
    n6 = n5.ConnectNew("n6")
    p.entrypoint = n1
    n4.condition = x.AddBinding("a", source_set=[], where=n2)
    n5.condition = y.AddBinding("a", source_set=[], where=n3)
    z_a = z.AddBinding("a", source_set=[], where=n1)
    # Impossible since we can only pass either n2 or n3.
    self.assertFalse(n6.HasCombination([z_a]))

  def test_conditions_block(self):
    p = cfg.Program()
    unreachable_node = p.NewCFGNode("unreachable_node")
    y = p.NewVariable()
    unsatisfiable_binding = y.AddBinding("2", source_set=[],
                                         where=unreachable_node)
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2", condition=unsatisfiable_binding)
    n3 = n2.ConnectNew("n3")
    x = p.NewVariable()
    b1 = x.AddBinding("1", source_set=[], where=n1)
    self.assertFalse(n3.HasCombination([b1]))
    n1.ConnectTo(n3)
    self.assertTrue(n3.HasCombination([b1]))
    self.assertFalse(n2.HasCombination([b1]))

  def test_conditions_multiple_paths(self):
    p = cfg.Program()
    unreachable_node = p.NewCFGNode("unreachable_node")
    y = p.NewVariable()
    unsatisfiable_binding = y.AddBinding("2", source_set=[],
                                         where=unreachable_node)
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2", condition=unsatisfiable_binding)
    n3 = n2.ConnectNew("n3")
    n4 = n2.ConnectNew("n4")
    n4.ConnectTo(n3)
    x = p.NewVariable()
    b1 = x.AddBinding("1", source_set=[], where=n1)
    self.assertFalse(n3.HasCombination([b1]))
    self.assertFalse(n2.HasCombination([b1]))

  def test_conditions_not_used_if_alternative_exist(self):
    p = cfg.Program()
    unreachable_node = p.NewCFGNode("unreachable_node")
    y = p.NewVariable()
    unsatisfiable_binding = y.AddBinding("2", source_set=[],
                                         where=unreachable_node)
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2", condition=unsatisfiable_binding)
    n3 = n2.ConnectNew("n3")
    x = p.NewVariable()
    b1 = x.AddBinding("1", source_set=[], where=n1)
    self.assertFalse(n3.HasCombination([b1]))

  def test_satisfiable_condition(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    x = p.NewVariable()
    x1 = x.AddBinding("1", source_set=[], where=n1)
    n2 = n1.ConnectNew("n2")
    y = p.NewVariable()
    y2 = y.AddBinding("2", source_set=[], where=n2)
    n3 = n2.ConnectNew("n3", condition=y2)
    n4 = n3.ConnectNew("n4")
    self.assertTrue(n4.HasCombination([x1]))

  def test_unsatisfiable_condition(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    x = p.NewVariable()
    x1 = x.AddBinding("1", source_set=[], where=n1)
    n2 = n1.ConnectNew("n2")
    x2 = x.AddBinding("2", source_set=[], where=n2)
    n3 = n2.ConnectNew("n3", condition=x2)
    n4 = n3.ConnectNew("n4")
    self.assertFalse(n4.HasCombination([x1]))

  def test_no_node_on_all_paths(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    y = p.NewVariable()
    y1 = y.AddBinding("y", source_set=[], where=n1)
    n3 = n2.ConnectNew("n3")
    n4 = n1.ConnectNew("n4")
    n5 = n4.ConnectNew("n5")
    n3.ConnectTo(n5)
    x = p.NewVariable()
    x1 = x.AddBinding("x", source_set=[], where=n2)
    n3.condition = x1
    n4.condition = x1
    self.assertTrue(n5.HasCombination([y1]))

  def test_condition_on_start_node(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    n3 = p.NewCFGNode("n3")
    a = p.NewVariable().AddBinding("a", source_set=[], where=n3)
    b = p.NewVariable().AddBinding("b", source_set=[], where=n1)
    n2.condition = a
    self.assertFalse(n2.HasCombination([b]))
    self.assertTrue(n1.HasCombination([b]))

  def test_condition_loop(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    n3 = p.NewCFGNode("n3")
    a = p.NewVariable().AddBinding("a")
    u1 = p.NewVariable().AddBinding("1", source_set=[], where=n3)
    p.NewVariable().AddBinding("2", source_set=[], where=n3)
    c = p.NewVariable().AddBinding("c", source_set=[u1], where=n1)
    a.AddOrigin(n2, [c])
    n2.condition = a
    self.assertFalse(n2.HasCombination([c]))

  def test_combinations(self):
    # n1------->n2
    #  |        |
    #  v        v
    # n3------->n4
    # [n2] x = a; y = a
    # [n3] x = b; y = b
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    n3 = n1.ConnectNew("n3")
    n4 = n2.ConnectNew("n4")
    n3.ConnectTo(n4)
    x = p.NewVariable()
    y = p.NewVariable()
    xa = x.AddBinding("a", source_set=[], where=n2)
    ya = y.AddBinding("a", source_set=[], where=n2)
    xb = x.AddBinding("b", source_set=[], where=n3)
    yb = y.AddBinding("b", source_set=[], where=n3)
    p.entrypoint = n1
    self.assertTrue(n4.HasCombination([xa, ya]))
    self.assertTrue(n4.HasCombination([xb, yb]))
    self.assertFalse(n4.HasCombination([xa, yb]))
    self.assertFalse(n4.HasCombination([xb, ya]))

  def test_conflicting(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    x = p.NewVariable()
    a = x.AddBinding("a", source_set=[], where=n1)
    b = x.AddBinding("b", source_set=[], where=n1)
    p.entrypoint = n1
    # At n1, x can either be a or b, but not both.
    self.assertTrue(n1.HasCombination([a]))
    self.assertTrue(n1.HasCombination([b]))
    self.assertFalse(n1.HasCombination([a, b]))

  def test_loop(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    n2.ConnectTo(n1)
    x = p.NewVariable()
    a = x.AddBinding("a")
    b = x.AddBinding("b")
    a.AddOrigin(n1, [b])
    b.AddOrigin(n2, [a])
    self.assertFalse(n2.HasCombination([b]))

  def test_one_step_simultaneous(self):
    # Like testSimultaneous, but woven through an additional node
    # n1->n2->n3
    # [n1] x = a or b
    # [n2] y = x
    # [n2] z = x
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    x = p.NewVariable()
    y = p.NewVariable()
    z = p.NewVariable()
    a = x.AddBinding("a", source_set=[], where=n1)
    b = x.AddBinding("b", source_set=[], where=n1)
    ya = y.AddBinding("ya", source_set=[a], where=n2)
    yb = y.AddBinding("yb", source_set=[b], where=n2)
    za = z.AddBinding("za", source_set=[a], where=n2)
    zb = z.AddBinding("zb", source_set=[b], where=n2)
    p.entrypoint = n1
    self.assertTrue(n2.HasCombination([ya, za]))
    self.assertTrue(n2.HasCombination([yb, zb]))
    self.assertFalse(n2.HasCombination([ya, zb]))
    self.assertFalse(n2.HasCombination([yb, za]))

  def test_conflicting_bindings(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    x = p.NewVariable()
    x_a = x.AddBinding("a", source_set=[], where=n1)
    x_b = x.AddBinding("b", source_set=[], where=n1)
    p.entrypoint = n1
    self.assertTrue(n1.HasCombination([x_a]))
    self.assertTrue(n1.HasCombination([x_b]))
    self.assertFalse(n1.HasCombination([x_a, x_b]))
    self.assertFalse(n2.HasCombination([x_a, x_b]))

  def test_mid_point(self):
    p = cfg.Program()
    x = p.NewVariable()
    y = p.NewVariable()
    n1 = p.NewCFGNode("n1")
    x1 = x.AddBinding("1", source_set=[], where=n1)
    y1 = y.AddBinding("1", source_set=[x1], where=n1)
    n2 = n1.ConnectNew("n2")
    x2 = x.AddBinding("2", source_set=[], where=n2)
    n3 = n2.ConnectNew("n3")
    self.assertTrue(n3.HasCombination([y1, x2]))
    self.assertTrue(n3.HasCombination([x2, y1]))

  def test_conditions_are_ordered(self):
    # The error case in this test is non-deterministic. The test tries to verify
    # that the list returned by _PathFinder.FindNodeBackwards is ordered from
    # child to parent.
    # The error case would be a random order or the reverse order.
    # To guarantee that this test is working go to FindNodeBackwards and reverse
    # the order of self._on_path before generating the returned list.
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    x1 = p.NewVariable().AddBinding("1", source_set=[], where=n1)
    n2 = n1.ConnectNew("n2", condition=p.NewVariable().AddBinding(
        "1", source_set=[], where=n1))
    n3 = n2.ConnectNew("n3", condition=p.NewVariable().AddBinding(
        "1", source_set=[], where=n2))
    n4 = n3.ConnectNew("n3", condition=p.NewVariable().AddBinding(
        "1", source_set=[], where=n3))
    # Strictly speaking n1, n2 and n3 would be enough to expose errors. n4 is
    # added to increase the chance of a failure if the order is random.
    self.assertTrue(n4.HasCombination([x1]))

  def test_same_node_origin(self):
    # [n1] x = a or b; y = x
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    x = p.NewVariable()
    y = p.NewVariable()
    xa = x.AddBinding("xa", source_set=[], where=n1)
    xb = x.AddBinding("xb", source_set=[], where=n1)
    ya = y.AddBinding("ya", source_set=[xa], where=n1)
    yb = y.AddBinding("yb", source_set=[xb], where=n1)
    p.entrypoint = n1
    self.assertTrue(n1.HasCombination([xa]))
    self.assertTrue(n1.HasCombination([xb]))
    self.assertTrue(n1.HasCombination([xa, ya]))
    self.assertTrue(n1.HasCombination([xb, yb]))
    # We don't check the other two combinations, because within one CFG node,
    # bindings are treated as having any order, so the other combinations
    # are possible, too:
    # n1.HasCombination([xa, yb]) == True (because x = b; y = x; x = a)
    # n1.HasCombination([xb, ya]) == True (because x = a; y = x; x = b)

  def test_new_variable(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = p.NewCFGNode("n2")
    x, y, z = "x", "y", "z"
    variable = p.NewVariable(bindings=[x, y],
                             source_set=[],
                             where=n1)
    variable.AddBinding(z, source_set=variable.bindings, where=n2)
    self.assertCountEqual([x, y, z], [v.data for v in variable.bindings])
    self.assertTrue(any(len(e.origins) for e in variable.bindings))
    # Test that non-list iterables can be passed to NewVariable.
    v2 = p.NewVariable((x, y), [], n1)
    self.assertCountEqual([x, y], [v.data for v in v2.bindings])
    v3 = p.NewVariable({x, y}, [], n1)
    self.assertCountEqual([x, y], [v.data for v in v3.bindings])
    v4 = p.NewVariable({x: y}, [], n1)
    self.assertCountEqual([x], [v.data for v in v4.bindings])

  def test_node_bindings(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("node1")
    n2 = n1.ConnectNew("node2")
    self.assertEqual(n1.name, "node1")
    self.assertEqual(n2.name, "node2")
    u = p.NewVariable()
    a1 = u.AddBinding(1, source_set=[], where=n1)
    a2 = u.AddBinding(2, source_set=[], where=n1)
    a3 = u.AddBinding(3, source_set=[], where=n1)
    a4 = u.AddBinding(4, source_set=[], where=n1)
    self.assertCountEqual([a1, a2, a3, a4], n1.bindings)

  def test_program(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    u1 = p.NewVariable()
    u2 = p.NewVariable()
    a11 = u1.AddBinding(11, source_set=[], where=n1)
    a12 = u1.AddBinding(12, source_set=[], where=n2)
    a21 = u2.AddBinding(21, source_set=[], where=n1)
    a22 = u2.AddBinding(22, source_set=[], where=n2)
    self.assertCountEqual([n1, n2], p.cfg_nodes)
    self.assertCountEqual([u1, u2], p.variables)
    self.assertCountEqual([a11, a21], n1.bindings)
    self.assertCountEqual([a12, a22], n2.bindings)
    self.assertEqual(p.next_variable_id, 2)

  def test_entry_point(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    x = p.NewVariable()
    a = x.AddBinding("a", source_set=[], where=n1)
    a = x.AddBinding("b", source_set=[], where=n2)
    p.entrypoint = n1
    self.assertTrue(n2.HasCombination([a]))

  def test_non_frozen_solving(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    x = p.NewVariable()
    a = x.AddBinding("a", source_set=[], where=n1)
    a = x.AddBinding("b", source_set=[], where=n2)
    p.entrypoint = n1
    self.assertTrue(n2.HasCombination([a]))

  def test_filter2(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = p.NewCFGNode("n2")
    n1.ConnectTo(n2)
    x = p.NewVariable()
    a = x.AddBinding("a", source_set=[], where=n2)
    p.entrypoint = n1
    self.assertEqual(x.Filter(n1), [])
    self.assertEqual(x.Filter(n2), [a])

  def test_hidden_conflict1(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    n3 = n1.ConnectNew("n3")
    x = p.NewVariable()
    y = p.NewVariable()
    z = p.NewVariable()
    x_a = x.AddBinding("a", source_set=[], where=n1)
    x_b = x.AddBinding("b", source_set=[], where=n1)
    y_a = y.AddBinding("a", source_set=[x_a], where=n1)
    y_b = y.AddBinding("b", source_set=[x_b], where=n2)
    z_ab1 = z.AddBinding("ab1", source_set=[x_a, x_b], where=n3)
    z_ab2 = z.AddBinding("ab2", source_set=[y_a, x_b], where=n3)
    z_ab3 = z.AddBinding("ab3", source_set=[y_b, x_a], where=n3)
    z_ab4 = z.AddBinding("ab4", source_set=[y_a, y_b], where=n3)
    p.entrypoint = n1
    self.assertFalse(n2.HasCombination([y_a, x_b]))
    self.assertFalse(n2.HasCombination([y_b, x_a]))
    self.assertFalse(n3.HasCombination([z_ab1]))
    self.assertFalse(n3.HasCombination([z_ab2]))
    self.assertFalse(n3.HasCombination([z_ab3]))
    self.assertFalse(n3.HasCombination([z_ab4]))

  def test_hidden_conflict2(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    x = p.NewVariable()
    y = p.NewVariable()
    x_a = x.AddBinding("a", source_set=[], where=n1)
    x_b = x.AddBinding("b", source_set=[], where=n1)
    y_b = y.AddBinding("b", source_set=[x_b], where=n1)
    p.entrypoint = n1
    self.assertFalse(n2.HasCombination([y_b, x_a]))

  def test_empty_binding(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    x = p.NewVariable()
    a = x.AddBinding("a")
    p.entrypoint = n1
    self.assertEqual(x.Filter(n1), [])
    self.assertEqual(x.Filter(n2), [])
    a.AddOrigin(n2, [])
    p.entrypoint = n1
    self.assertEqual(x.Filter(n1), [])
    self.assertEqual(x.Filter(n2), [a])
    a.AddOrigin(n1, [a])
    p.entrypoint = n1
    self.assertEqual(x.Filter(n1), [a])
    self.assertEqual(x.Filter(n2), [a])

  def test_assign_to_new(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    n3 = n2.ConnectNew("n3")
    x = p.NewVariable()
    ax = x.AddBinding("a", source_set=[], where=n1)
    y = ax.AssignToNewVariable(n2)
    ay, = y.bindings
    z = y.AssignToNewVariable(n3)
    az, = z.bindings
    self.assertEqual([v.data for v in y.bindings], ["a"])
    self.assertEqual([v.data for v in z.bindings], ["a"])
    p.entrypoint = n1
    self.assertTrue(n1.HasCombination([ax]))
    self.assertTrue(n2.HasCombination([ax, ay]))
    self.assertTrue(n3.HasCombination([ax, ay, az]))
    self.assertFalse(n1.HasCombination([ax, ay]))
    self.assertFalse(n2.HasCombination([ax, ay, az]))

  def test_assign_to_new_no_node(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    x = p.NewVariable()
    ax = x.AddBinding("a", source_set=[], where=n1)
    y = ax.AssignToNewVariable()
    z = x.AssignToNewVariable()
    ox, = x.bindings[0].origins
    oy, = y.bindings[0].origins
    oz, = z.bindings[0].origins
    self.assertEqual(ox, oy, oz)

  def test_paste_variable(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    x = p.NewVariable()
    ax = x.AddBinding("a", source_set=[], where=n1)
    bx = x.AddBinding("b", source_set=[], where=n1)
    y = p.NewVariable()
    y.PasteVariable(x, n2)
    ay, by = y.bindings
    self.assertEqual([v.data for v in x.bindings], ["a", "b"])
    self.assertEqual([v.data for v in y.bindings], ["a", "b"])
    p.entrypoint = n1
    self.assertTrue(n1.HasCombination([ax]))
    self.assertTrue(n1.HasCombination([bx]))
    self.assertFalse(n1.HasCombination([ay]))
    self.assertFalse(n1.HasCombination([by]))
    self.assertTrue(n2.HasCombination([ay]))
    self.assertTrue(n2.HasCombination([by]))

  def test_paste_at_same_node(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    x = p.NewVariable()
    x.AddBinding("a", source_set=[], where=n1)
    x.AddBinding("b", source_set=[], where=n1)
    y = p.NewVariable()
    y.PasteVariable(x, n1)
    ay, _ = y.bindings
    self.assertEqual([v.data for v in x.bindings], ["a", "b"])
    self.assertEqual([v.data for v in y.bindings], ["a", "b"])
    o, = ay.origins
    self.assertCountEqual([set()], o.source_sets)
    o, = ay.origins
    self.assertCountEqual([set()], o.source_sets)

  def test_paste_with_additional_sources(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    x = p.NewVariable()
    y = p.NewVariable()
    z = p.NewVariable()
    ax = x.AddBinding("a", source_set=[], where=n1)
    by = y.AddBinding("b", source_set=[], where=n1)
    z.PasteVariable(x, n2, {by})
    az, = z.bindings
    origin, = az.origins
    source_set, = origin.source_sets
    self.assertSetEqual(source_set, {ax, by})

  def test_paste_at_same_node_with_additional_sources(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    x = p.NewVariable()
    y = p.NewVariable()
    z = p.NewVariable()
    _ = x.AddBinding("a", source_set=[], where=n1)
    by = y.AddBinding("b", source_set=[], where=n1)
    z.PasteVariable(x, n1, {by})
    az, = z.bindings
    origin, = az.origins
    source_set, = origin.source_sets
    self.assertSetEqual(source_set, {by})

  def test_paste_binding(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    x = p.NewVariable()
    ax = x.AddBinding("a", source_set=[], where=n1)
    y = p.NewVariable()
    y.PasteBinding(ax)
    self.assertEqual(x.data, y.data)

  def test_id(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = p.NewCFGNode("n2")
    x = p.NewVariable()
    y = p.NewVariable()
    self.assertIsInstance(x.id, int)
    self.assertIsInstance(y.id, int)
    self.assertLess(x.id, y.id)
    self.assertIsInstance(n1.id, int)
    self.assertIsInstance(n2.id, int)
    self.assertLess(n1.id, n2.id)

  def test_prune(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    n3 = n2.ConnectNew("n3")
    n4 = n3.ConnectNew("n4")
    n1.ConnectTo(n4)
    x = p.NewVariable()
    x.AddBinding(1, [], n1)
    x.AddBinding(2, [], n2)
    x.AddBinding(3, [], n3)
    self.assertCountEqual([1], [v.data for v in x.Bindings(n1)])
    self.assertCountEqual([2], [v.data for v in x.Bindings(n2)])
    self.assertCountEqual([3], [v.data for v in x.Bindings(n3)])
    self.assertCountEqual([1, 3], [v.data for v in x.Bindings(n4)])
    self.assertCountEqual([1], x.Data(n1))
    self.assertCountEqual([2], x.Data(n2))
    self.assertCountEqual([3], x.Data(n3))
    self.assertCountEqual([1, 3], x.Data(n4))

  def test_prune_two_origins(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = p.NewCFGNode("n2")
    n3 = p.NewCFGNode("n2")
    n1.ConnectTo(n3)
    n2.ConnectTo(n3)
    x = p.NewVariable()
    b = x.AddBinding(1)
    b.AddOrigin(source_set=[], where=n1)
    b.AddOrigin(source_set=[], where=n2)
    self.assertEqual(len([v.data for v in x.Bindings(n3)]), 1)

  def test_hidden_conflict3(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    z = p.NewVariable()
    z_a = z.AddBinding("a", source_set=[], where=n1)
    z_b = z.AddBinding("b", source_set=[], where=n1)
    goals = []
    for _ in range(5):
      var = p.NewVariable()
      v = var.AddBinding(".")
      v.AddOrigin(source_set=[z_a], where=n1)
      v.AddOrigin(source_set=[z_b], where=n1)
      goals.append(v)
    x = p.NewVariable()
    x_b = x.AddBinding("a", source_set=[z_b], where=n1)
    self.assertTrue(n2.HasCombination(goals + [x_b]))

  def test_conflict_with_condition(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    z = p.NewVariable()
    z_a = z.AddBinding("a", source_set=[], where=n1)
    z_b = z.AddBinding("b", source_set=[], where=n1)
    n1.condition = z_b
    goals = []
    for _ in range(5):
      var = p.NewVariable()
      v = var.AddBinding(".")
      v.AddOrigin(source_set=[z_a], where=n1)
      v.AddOrigin(source_set=[z_b], where=n1)
      goals.append(v)
    self.assertTrue(n2.HasCombination(goals))

  def test_variable_properties(self):
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    n2 = p.NewCFGNode("n2")
    n3 = p.NewCFGNode("n3")
    v = p.NewVariable()
    v.AddBinding("a", source_set=[], where=n1)
    v.AddBinding("b", source_set=[], where=n2)
    v.AddBinding("c", source_set=[], where=n3)
    self.assertCountEqual(v.data, ["a", "b", "c"])
    self.assertCountEqual(v.bindings, v.Bindings(None))
    self.assertEqual(p, v.program)

  def test_add_binding_iterables(self):
    # source_set in Pytype is at times a tuple, list or set. They're all
    # converted to SourceSets (essentially frozensets) when added to an Origin.
    # This is more of a behavioral test than a specification test.
    p = cfg.Program()
    n1 = p.NewCFGNode("n1")
    x = p.NewVariable()
    x.AddBinding("a", source_set=(), where=n1)
    x.AddBinding("b", source_set=[], where=n1)
    x.AddBinding("c", source_set=set(), where=n1)

  def test_calls_with_none(self):
    # Several parts of the Python API have None as a default value for
    # parameters. Make sure the C++ API can # also take None for those
    # functions. These are mostly smoke tests.
    p = cfg.Program()
    n1 = p.NewCFGNode()
    n2 = p.NewCFGNode(None)
    self.assertEqual(n1.name, n2.name)
    v1 = p.NewVariable()
    v2 = p.NewVariable(None)
    self.assertEqual(v1.bindings, [])
    self.assertEqual(v1.bindings, v2.bindings)
    v3 = p.NewVariable(None, None, None)
    self.assertEqual(v1.bindings, v3.bindings)
    n3 = n1.ConnectNew()
    n4 = n1.ConnectNew(None)
    self.assertEqual(n3.name, n4.name)
    n5 = n1.ConnectNew(None, None)
    self.assertEqual(n1.condition, n5.condition)
    av = v1.AddBinding("a")
    bv = v1.AddBinding("b", None, None)
    v3 = av.AssignToNewVariable()
    v1.PasteVariable(v2)
    v1.PasteVariable(v2, None)
    v1.PasteVariable(v2, None, None)
    v2.PasteBinding(bv)
    v2.PasteBinding(bv, None)
    v2.PasteBinding(bv, None, None)

  def test_program_default_data(self):
    # Basic sanity check to make sure Program.default_data works.
    p = cfg.Program()
    self.assertIsNone(p.default_data)
    p.default_data = 1
    self.assertEqual(p.default_data, 1)

  def test_condition_conflict(self):
    # v1 = x or y or z  # node_in
    # condition = v1 is x  # node_in
    # if condition:  # node_if
    #   assert v1 is x
    #   assert v1 is not y and v1 is not z
    # else:  # node_else
    #   assert v1 is not x
    #   assert v1 is y or v1 is z
    p = cfg.Program()
    node_in = p.NewCFGNode("node_in")
    v1 = p.NewVariable()
    bx = v1.AddBinding("x", [], node_in)
    by = v1.AddBinding("y", [], node_in)
    bz = v1.AddBinding("z", [], node_in)
    condition_true = p.NewVariable().AddBinding(True, [bx], node_in)
    condition_false = condition_true.variable.AddBinding(False, [by], node_in)
    condition_false.AddOrigin(node_in, [bz])
    b_if = p.NewVariable().AddBinding("if", [condition_true], node_in)
    b_else = p.NewVariable().AddBinding("else", [condition_false], node_in)
    node_if = node_in.ConnectNew("node_if", b_if)
    node_else = node_in.ConnectNew("node_else", b_else)
    self.assertTrue(b_if.IsVisible(node_if))
    self.assertFalse(b_else.IsVisible(node_if))
    self.assertFalse(b_if.IsVisible(node_else))
    self.assertTrue(b_else.IsVisible(node_else))

  def test_block_condition(self):
    # v1 = x or y or z  # node_in
    # if v1 is x:  # node_if
    #   v1 = w  # node_block
    # else: ...  # node_else
    # assert v1 is not x  # node_out
    p = cfg.Program()
    node_in = p.NewCFGNode("node_in")
    v1 = p.NewVariable()
    bx = v1.AddBinding("x", [], node_in)
    by = v1.AddBinding("y", [], node_in)
    bz = v1.AddBinding("z", [], node_in)
    b_if = p.NewVariable().AddBinding("if", [bx], node_in)
    b_else = b_if.variable.AddBinding("else", [by], node_in)
    b_else.AddOrigin(node_in, [bz])
    node_if = node_in.ConnectNew("node_if", b_if)
    node_else = node_in.ConnectNew("node_else", b_else)
    node_block = node_if.ConnectNew("node_block")
    v1.AddBinding("w", [], node_block)
    node_out = node_block.ConnectNew("node_out")
    node_else.ConnectTo(node_out)
    b_out = p.NewVariable().AddBinding("x", [bx], node_out)
    self.assertFalse(b_out.IsVisible(node_out))

  def test_strict(self):
    # Tests the existence of the strict keyword (but not its behavior).
    p = cfg.Program()
    node = p.NewCFGNode("root")
    v = p.NewVariable()
    b = v.AddBinding("x", [], node)
    self.assertEqual(v.Filter(node, strict=False), [b])
    self.assertEqual(v.FilteredData(node, strict=False), [b.data])
    self.assertEqual(v.Bindings(node, strict=False), [b])


if __name__ == "__main__":
  unittest.main()



from bdb import Bdb
from cmd import Cmd
from types import FrameType, TracebackType
from typing import IO, Any, Callable, ClassVar, Dict, Iterable, List, Mapping, Tuple, TypeVar

_T = TypeVar("_T")

line_prefix: str  # undocumented

class Restart(Exception): ...

def run(statement: str, globals: Dict[str, Any] | None = ..., locals: Mapping[str, Any] | None = ...) -> None: ...
def runeval(expression: str, globals: Dict[str, Any] | None = ..., locals: Mapping[str, Any] | None = ...) -> Any: ...
def runctx(statement: str, globals: Dict[str, Any], locals: Mapping[str, Any]) -> None: ...
def runcall(func: Callable[..., _T], *args: Any, **kwds: Any) -> _T | None: ...
def set_trace() -> None: ...
def post_mortem(t: TracebackType | None = ...) -> None: ...
def pm() -> None: ...

class Pdb(Bdb, Cmd):
    # Everything here is undocumented, except for __init__

    commands_resuming: ClassVar[List[str]]

    aliases: Dict[str, str]
    mainpyfile: str
    _wait_for_mainpyfile: bool
    rcLines: List[str]
    commands: Dict[int, List[str]]
    commands_doprompt: Dict[int, bool]
    commands_silent: Dict[int, bool]
    commands_defining: bool
    commands_bnum: int | None
    lineno: int | None
    stack: List[Tuple[FrameType, int]]
    curindex: int
    curframe: FrameType | None
    curframe_locals: Mapping[str, Any]
    def __init__(
        self, completekey: str = ..., stdin: IO[str] | None = ..., stdout: IO[str] | None = ..., skip: Iterable[str] | None = ...
    ) -> None: ...
    def forget(self) -> None: ...
    def setup(self, f: FrameType | None, tb: TracebackType | None) -> None: ...
    def execRcLines(self) -> None: ...
    def bp_commands(self, frame: FrameType) -> bool: ...
    def interaction(self, frame: FrameType | None, traceback: TracebackType | None) -> None: ...
    def displayhook(self, obj: object) -> None: ...
    def handle_command_def(self, line: str) -> bool: ...
    def defaultFile(self) -> str: ...
    def lineinfo(self, identifier: str) -> Tuple[None, None, None] | Tuple[str, str, int]: ...
    def checkline(self, filename: str, lineno: int) -> int: ...
    def _getval(self, arg: str) -> object: ...
    def print_stack_trace(self) -> None: ...
    def print_stack_entry(self, frame_lineno: Tuple[FrameType, int], prompt_prefix: str = ...) -> None: ...
    def lookupmodule(self, filename: str) -> str | None: ...
    def _runscript(self, filename: str) -> None: ...
    def do_commands(self, arg: str) -> bool | None: ...
    def do_break(self, arg: str, temporary: bool = ...) -> bool | None: ...
    def do_tbreak(self, arg: str) -> bool | None: ...
    def do_enable(self, arg: str) -> bool | None: ...
    def do_disable(self, arg: str) -> bool | None: ...
    def do_condition(self, arg: str) -> bool | None: ...
    def do_ignore(self, arg: str) -> bool | None: ...
    def do_clear(self, arg: str) -> bool | None: ...
    def do_where(self, arg: str) -> bool | None: ...
    def do_up(self, arg: str) -> bool | None: ...
    def do_down(self, arg: str) -> bool | None: ...
    def do_until(self, arg: str) -> bool | None: ...
    def do_step(self, arg: str) -> bool | None: ...
    def do_next(self, arg: str) -> bool | None: ...
    def do_run(self, arg: str) -> bool | None: ...
    def do_return(self, arg: str) -> bool | None: ...
    def do_continue(self, arg: str) -> bool | None: ...
    def do_jump(self, arg: str) -> bool | None: ...
    def do_debug(self, arg: str) -> bool | None: ...
    def do_quit(self, arg: str) -> bool | None: ...
    def do_EOF(self, arg: str) -> bool | None: ...
    def do_args(self, arg: str) -> bool | None: ...
    def do_retval(self, arg: str) -> bool | None: ...
    def do_p(self, arg: str) -> bool | None: ...
    def do_pp(self, arg: str) -> bool | None: ...
    def do_list(self, arg: str) -> bool | None: ...
    def do_whatis(self, arg: str) -> bool | None: ...
    def do_alias(self, arg: str) -> bool | None: ...
    def do_unalias(self, arg: str) -> bool | None: ...
    def do_help(self, arg: str) -> bool | None: ...
    do_b = do_break
    do_cl = do_clear
    do_w = do_where
    do_bt = do_where
    do_u = do_up
    do_d = do_down
    do_unt = do_until
    do_s = do_step
    do_n = do_next
    do_restart = do_run
    do_r = do_return
    do_c = do_continue
    do_cont = do_continue
    do_j = do_jump
    do_q = do_quit
    do_exit = do_quit
    do_a = do_args
    do_rv = do_retval
    do_l = do_list
    do_h = do_help
    def help_exec(self) -> None: ...
    def help_pdb(self) -> None: ...
    def help_help(self) -> None: ...
    def help_h(self) -> None: ...
    def help_where(self) -> None: ...
    def help_w(self) -> None: ...
    def help_down(self) -> None: ...
    def help_d(self) -> None: ...
    def help_up(self) -> None: ...
    def help_u(self) -> None: ...
    def help_break(self) -> None: ...
    def help_b(self) -> None: ...
    def help_clear(self) -> None: ...
    def help_cl(self) -> None: ...
    def help_tbreak(self) -> None: ...
    def help_enable(self) -> None: ...
    def help_disable(self) -> None: ...
    def help_ignore(self) -> None: ...
    def help_condition(self) -> None: ...
    def help_step(self) -> None: ...
    def help_s(self) -> None: ...
    def help_until(self) -> None: ...
    def help_unt(self) -> None: ...
    def help_next(self) -> None: ...
    def help_n(self) -> None: ...
    def help_return(self) -> None: ...
    def help_r(self) -> None: ...
    def help_continue(self) -> None: ...
    def help_cont(self) -> None: ...
    def help_c(self) -> None: ...
    def help_jump(self) -> None: ...
    def help_j(self) -> None: ...
    def help_debug(self) -> None: ...
    def help_list(self) -> None: ...
    def help_l(self) -> None: ...
    def help_args(self) -> None: ...
    def help_a(self) -> None: ...
    def help_p(self) -> None: ...
    def help_pp(self) -> None: ...
    def help_run(self) -> None: ...
    def help_quit(self) -> None: ...
    def help_q(self) -> None: ...
    def help_whatis(self) -> None: ...
    def help_EOF(self) -> None: ...
    def help_alias(self) -> None: ...
    def help_unalias(self) -> None: ...
    def help_commands(self) -> None: ...
    help_bt = help_w
    help_restart = help_run
    help_exit = help_q

# undocumented

def find_function(funcname: str, filename: str) -> Tuple[str, str, int] | None: ...
def main() -> None: ...
def help() -> None: ...

class _rstr(str):
    def __repr__(self) -> _rstr: ...



from _typeshed import SupportsRead
from typing import Any, overload
from typing_extensions import Literal
from xml.etree.ElementTree import Element

@overload
def parse(
    doc: str | bytes | SupportsRead[str] | SupportsRead[bytes],
    treebuilder: Literal["etree"] = ...,
    namespaceHTMLElements: bool = ...,
    **kwargs,
) -> Element: ...
@overload
def parse(
    doc: str | bytes | SupportsRead[str] | SupportsRead[bytes], treebuilder: str, namespaceHTMLElements: bool = ..., **kwargs
): ...
def parseFragment(doc, container: str = ..., treebuilder: str = ..., namespaceHTMLElements: bool = ..., **kwargs): ...
def method_decorator_metaclass(function): ...

class HTMLParser:
    strict: Any
    tree: Any
    errors: Any
    phases: Any
    def __init__(
        self, tree: Any | None = ..., strict: bool = ..., namespaceHTMLElements: bool = ..., debug: bool = ...
    ) -> None: ...
    firstStartTag: bool
    log: Any
    compatMode: str
    innerHTML: Any
    phase: Any
    lastPhase: Any
    beforeRCDataPhase: Any
    framesetOK: bool
    def reset(self) -> None: ...
    @property
    def documentEncoding(self) -> str | None: ...
    def isHTMLIntegrationPoint(self, element) -> bool: ...
    def isMathMLTextIntegrationPoint(self, element) -> bool: ...
    def mainLoop(self) -> None: ...
    def parse(self, stream, scripting: bool = ..., **kwargs): ...
    def parseFragment(self, stream, *args, **kwargs): ...
    def parseError(self, errorcode: str = ..., datavars: Any | None = ...) -> None: ...
    def adjustMathMLAttributes(self, token) -> None: ...
    def adjustSVGAttributes(self, token) -> None: ...
    def adjustForeignAttributes(self, token) -> None: ...
    def reparseTokenNormal(self, token) -> None: ...
    def resetInsertionMode(self) -> None: ...
    originalPhase: Any
    def parseRCDataRawtext(self, token, contentType) -> None: ...

def getPhases(debug): ...
def adjust_attributes(token, replacements) -> None: ...
def impliedTagToken(name, type: str = ..., attributes: Any | None = ..., selfClosing: bool = ...): ...

class ParseError(Exception): ...



from datetime import timedelta
from typing import Any, Tuple

from .base import Component
from .behavior import Behavior

DATENAMES: Tuple[str, ...]
RULENAMES: Tuple[str, ...]
DATESANDRULES: Tuple[str, ...]
PRODID: str
WEEKDAYS: Tuple[str, ...]
FREQUENCIES: Tuple[str, ...]
zeroDelta: timedelta
twoHours: timedelta

def toUnicode(s: str | bytes) -> str: ...
def registerTzid(tzid, tzinfo) -> None: ...
def getTzid(tzid, smart: bool = ...): ...

utc: Any  # dateutil.tz.tz.tzutc

class TimezoneComponent(Component):
    isNative: bool
    behavior: Any
    tzinfo: Any
    name: str
    useBegin: bool
    def __init__(self, tzinfo: Any | None = ..., *args, **kwds) -> None: ...
    @classmethod
    def registerTzinfo(cls, tzinfo): ...
    def gettzinfo(self): ...
    tzid: Any
    daylight: Any
    standard: Any
    def settzinfo(self, tzinfo, start: int = ..., end: int = ...): ...
    normal_attributes: Any
    @staticmethod
    def pickTzid(tzinfo, allowUTC: bool = ...): ...
    def prettyPrint(self, level, tabwidth) -> None: ...  # type: ignore[override]

class RecurringComponent(Component):
    isNative: bool
    def __init__(self, *args, **kwds) -> None: ...
    def getrruleset(self, addRDate: bool = ...): ...
    def setrruleset(self, rruleset): ...
    rruleset: Any
    def __setattr__(self, name, value) -> None: ...

class TextBehavior(Behavior):
    base64string: str
    @classmethod
    def decode(cls, line) -> None: ...
    @classmethod
    def encode(cls, line) -> None: ...

class VCalendarComponentBehavior(Behavior):
    defaultBehavior: Any
    isComponent: bool

class RecurringBehavior(VCalendarComponentBehavior):
    hasNative: bool
    @staticmethod
    def transformToNative(obj): ...
    @staticmethod
    def transformFromNative(obj): ...
    @staticmethod
    def generateImplicitParameters(obj) -> None: ...

class DateTimeBehavior(Behavior):
    hasNative: bool
    @staticmethod
    def transformToNative(obj): ...
    @classmethod
    def transformFromNative(cls, obj): ...

class UTCDateTimeBehavior(DateTimeBehavior):
    forceUTC: bool

class DateOrDateTimeBehavior(Behavior):
    hasNative: bool
    @staticmethod
    def transformToNative(obj): ...
    @staticmethod
    def transformFromNative(obj): ...

class MultiDateBehavior(Behavior):
    hasNative: bool
    @staticmethod
    def transformToNative(obj): ...
    @staticmethod
    def transformFromNative(obj): ...

class MultiTextBehavior(Behavior):
    listSeparator: str
    @classmethod
    def decode(cls, line) -> None: ...
    @classmethod
    def encode(cls, line) -> None: ...

class SemicolonMultiTextBehavior(MultiTextBehavior):
    listSeparator: str

class VCalendar2_0(VCalendarComponentBehavior):
    name: str
    description: str
    versionString: str
    sortFirst: Any
    knownChildren: Any
    @classmethod
    def generateImplicitParameters(cls, obj) -> None: ...
    @classmethod
    def serialize(cls, obj, buf, lineLength, validate: bool = ...): ...

class VTimezone(VCalendarComponentBehavior):
    name: str
    hasNative: bool
    description: str
    sortFirst: Any
    knownChildren: Any
    @classmethod
    def validate(cls, obj, raiseException, *args): ...
    @staticmethod
    def transformToNative(obj): ...
    @staticmethod
    def transformFromNative(obj): ...

class TZID(Behavior): ...

class DaylightOrStandard(VCalendarComponentBehavior):
    hasNative: bool
    knownChildren: Any

class VEvent(RecurringBehavior):
    name: str
    sortFirst: Any
    description: str
    knownChildren: Any
    @classmethod
    def validate(cls, obj, raiseException, *args): ...

class VTodo(RecurringBehavior):
    name: str
    description: str
    knownChildren: Any
    @classmethod
    def validate(cls, obj, raiseException, *args): ...

class VJournal(RecurringBehavior):
    name: str
    knownChildren: Any

class VFreeBusy(VCalendarComponentBehavior):
    name: str
    description: str
    sortFirst: Any
    knownChildren: Any

class VAlarm(VCalendarComponentBehavior):
    name: str
    description: str
    knownChildren: Any
    @staticmethod
    def generateImplicitParameters(obj) -> None: ...
    @classmethod
    def validate(cls, obj, raiseException, *args): ...

class VAvailability(VCalendarComponentBehavior):
    name: str
    description: str
    sortFirst: Any
    knownChildren: Any
    @classmethod
    def validate(cls, obj, raiseException, *args): ...

class Available(RecurringBehavior):
    name: str
    sortFirst: Any
    description: str
    knownChildren: Any
    @classmethod
    def validate(cls, obj, raiseException, *args): ...

class Duration(Behavior):
    name: str
    hasNative: bool
    @staticmethod
    def transformToNative(obj): ...
    @staticmethod
    def transformFromNative(obj): ...

class Trigger(Behavior):
    name: str
    description: str
    hasNative: bool
    forceUTC: bool
    @staticmethod
    def transformToNative(obj): ...
    @staticmethod
    def transformFromNative(obj): ...

class PeriodBehavior(Behavior):
    hasNative: bool
    @staticmethod
    def transformToNative(obj): ...
    @classmethod
    def transformFromNative(cls, obj): ...

class FreeBusy(PeriodBehavior):
    name: str
    forceUTC: bool

class RRule(Behavior): ...

utcDateTimeList: Any
dateTimeOrDateList: Any
textList: Any

def numToDigits(num, places): ...
def timedeltaToString(delta): ...
def timeToString(dateOrDateTime): ...
def dateToString(date): ...
def dateTimeToString(dateTime, convertToUTC: bool = ...): ...
def deltaToOffset(delta): ...
def periodToString(period, convertToUTC: bool = ...): ...
def isDuration(s): ...
def stringToDate(s): ...
def stringToDateTime(s, tzinfo: Any | None = ...): ...

escapableCharList: str

def stringToTextValues(s, listSeparator: str = ..., charList: Any | None = ..., strict: bool = ...): ...
def stringToDurations(s, strict: bool = ...): ...
def parseDtstart(contentline, allowSignatureMismatch: bool = ...): ...
def stringToPeriod(s, tzinfo: Any | None = ...): ...
def getTransition(transitionTo, year, tzinfo): ...
def tzinfo_eq(tzinfo1, tzinfo2, startYear: int = ..., endYear: int = ...): ...



from typing import Any
def f1(x):
    # type: (Any) -> Union[int, str]
    return 1
def f1(x):
    # type: (Any) -> Union[int, str]
    return 'foo'

def f2(x):
    # type: (Any) -> None
    pass
def f2(x,y):
    pass

def f3(x):
    # type: (int) -> int
    return 1+x
def f3(x):
    # type: (int) -> int
    return 'asd'+x



from typing import Dict, List, Sequence

cmp_op: Sequence[str]
hasconst: List[int]
hasname: List[int]
hasjrel: List[int]
hasjabs: List[int]
haslocal: List[int]
hascompare: List[int]
hasfree: List[int]
opname: List[str]

opmap: Dict[str, int]
HAVE_ARGUMENT: int
EXTENDED_ARG: int



from abc import abstractmethod
from email.contentmanager import ContentManager
from email.errors import MessageDefect
from email.header import Header
from email.message import Message
from typing import Any, Callable

class Policy:
    max_line_length: int | None
    linesep: str
    cte_type: str
    raise_on_defect: bool
    mange_from: bool
    def __init__(self, **kw: Any) -> None: ...
    def clone(self, **kw: Any) -> Policy: ...
    def handle_defect(self, obj: Message, defect: MessageDefect) -> None: ...
    def register_defect(self, obj: Message, defect: MessageDefect) -> None: ...
    def header_max_count(self, name: str) -> int | None: ...
    @abstractmethod
    def header_source_parse(self, sourcelines: list[str]) -> tuple[str, str]: ...
    @abstractmethod
    def header_store_parse(self, name: str, value: str) -> tuple[str, str]: ...
    @abstractmethod
    def header_fetch_parse(self, name: str, value: str) -> str: ...
    @abstractmethod
    def fold(self, name: str, value: str) -> str: ...
    @abstractmethod
    def fold_binary(self, name: str, value: str) -> bytes: ...

class Compat32(Policy):
    def header_source_parse(self, sourcelines: list[str]) -> tuple[str, str]: ...
    def header_store_parse(self, name: str, value: str) -> tuple[str, str]: ...
    def header_fetch_parse(self, name: str, value: str) -> str | Header: ...  # type: ignore[override]
    def fold(self, name: str, value: str) -> str: ...
    def fold_binary(self, name: str, value: str) -> bytes: ...

compat32: Compat32

class EmailPolicy(Policy):
    utf8: bool
    refold_source: str
    header_factory: Callable[[str, str], str]
    content_manager: ContentManager
    def header_source_parse(self, sourcelines: list[str]) -> tuple[str, str]: ...
    def header_store_parse(self, name: str, value: str) -> tuple[str, str]: ...
    def header_fetch_parse(self, name: str, value: str) -> str: ...
    def fold(self, name: str, value: str) -> str: ...
    def fold_binary(self, name: str, value: str) -> bytes: ...

default: EmailPolicy
SMTP: EmailPolicy
SMTPUTF8: EmailPolicy
HTTP: EmailPolicy
strict: EmailPolicy



"""Tests for environment.py."""

import os

from pytype import file_utils
from pytype.tools import environment
import unittest


class TestComputePythonPath(unittest.TestCase):
  """Tests for environment.compute_pythonpath."""

  def test_script_path(self):
    with file_utils.Tempdir() as d:
      f = d.create_file('foo.py')
      self.assertSequenceEqual(environment.compute_pythonpath([f]), [d.path])

  def test_module_path(self):
    with file_utils.Tempdir() as d:
      d.create_file('__init__.py')
      f = d.create_file('foo.py')
      self.assertSequenceEqual(environment.compute_pythonpath([f]),
                               [os.path.dirname(d.path)])

  def test_subpackage(self):
    with file_utils.Tempdir() as d:
      d.create_file('__init__.py')
      d.create_file('d/__init__.py')
      f = d.create_file('d/foo.py')
      self.assertSequenceEqual(environment.compute_pythonpath([f]),
                               [os.path.dirname(d.path)])

  def test_multiple_paths(self):
    with file_utils.Tempdir() as d:
      f1 = d.create_file('d1/foo.py')
      f2 = d.create_file('d2/foo.py')
      self.assertSequenceEqual(
          environment.compute_pythonpath([f1, f2]),
          [os.path.join(d.path, 'd2'), os.path.join(d.path, 'd1')])

  def test_sort(self):
    with file_utils.Tempdir() as d:
      f1 = d.create_file('d1/foo.py')
      f2 = d.create_file('d1/d2/foo.py')
      f3 = d.create_file('d1/d2/d3/foo.py')
      path = [os.path.join(d.path, 'd1', 'd2', 'd3'),
              os.path.join(d.path, 'd1', 'd2'),
              os.path.join(d.path, 'd1')]
      self.assertSequenceEqual(
          environment.compute_pythonpath([f1, f2, f3]), path)
      self.assertSequenceEqual(
          environment.compute_pythonpath([f3, f2, f1]), path)


class TestDoXOrDie(unittest.TestCase):
  """Tests for {do_x}_or_die() methods.

  Since whether these functions complete successfully depends on one's
  particular environment, these tests allow either succeeding or raising
  SystemExit. Any other exception will cause a test failure.
  """

  def _test(self, method, *args):
    try:
      method(*args)
    except SystemExit:
      pass

  def test_pytype(self):
    self._test(environment.check_pytype_or_die)

  def test_python_exe(self):
    self._test(environment.check_python_exe_or_die, 3.0)

  def test_typeshed(self):
    self._test(environment.initialize_typeshed_or_die)


if __name__ == '__main__':
  unittest.main()



"""Tests for the additional CFG utilities."""

import itertools

from pytype.typegraph import cfg
from pytype.typegraph import cfg_utils
import unittest


class CFGUtilTest(unittest.TestCase):
  """Test CFG utilities."""

  def test_merge_zero_variables(self):
    p = cfg.Program()
    n0 = p.NewCFGNode("n0")
    self.assertIsInstance(cfg_utils.merge_variables(p, n0, []), cfg.Variable)

  def test_merge_one_variable(self):
    p = cfg.Program()
    n0 = p.NewCFGNode("n0")
    u = p.NewVariable([0], [], n0)
    self.assertIsNot(cfg_utils.merge_variables(p, n0, [u]), u)
    self.assertIsNot(cfg_utils.merge_variables(p, n0, [u, u, u]), u)
    self.assertCountEqual(cfg_utils.merge_variables(p, n0, [u, u, u]).data,
                          u.data)

  def test_merge_variables(self):
    p = cfg.Program()
    n0, n1, n2 = p.NewCFGNode("n0"), p.NewCFGNode("n1"), p.NewCFGNode("n2")
    u = p.NewVariable()
    u1 = u.AddBinding(0, source_set=[], where=n0)
    v = p.NewVariable()
    v.AddBinding(1, source_set=[], where=n1)
    v.AddBinding(2, source_set=[], where=n1)
    w = p.NewVariable()
    w.AddBinding(1, source_set=[u1], where=n1)
    w.AddBinding(3, source_set=[], where=n1)
    vw = cfg_utils.merge_variables(p, n2, [v, w])
    self.assertCountEqual(vw.data, [1, 2, 3])
    val1, = [v for v in vw.bindings if v.data == 1]
    self.assertTrue(val1.HasSource(u1))

  def test_merge_bindings(self):
    p = cfg.Program()
    n0 = p.NewCFGNode("n0")
    u = p.NewVariable()
    u1 = u.AddBinding("1", source_set=[], where=n0)
    v2 = u.AddBinding("2", source_set=[], where=n0)
    w1 = cfg_utils.merge_bindings(p, None, [u1, v2])
    w2 = cfg_utils.merge_bindings(p, n0, [u1, v2])
    self.assertCountEqual(w1.data, ["1", "2"])
    self.assertCountEqual(w2.data, ["1", "2"])


class DummyValue:
  """A class with a 'parameters' function, for testing cartesian products."""

  def __init__(self, index):
    self.index = index
    self._parameters = []

  def set_parameters(self, parameters):
    self._parameters = parameters

  def unique_parameter_values(self):
    return [param.bindings for param in self._parameters]

  def __repr__(self):
    return "x%d" % self.index


class VariableProductTest(unittest.TestCase):
  """Test variable-product utilities."""

  def setUp(self):
    super().setUp()
    self.prog = cfg.Program()
    self.current_location = self.prog.NewCFGNode()

  def test_complexity_limit(self):
    limit = cfg_utils.ComplexityLimit(5)
    limit.inc()
    limit.inc(2)
    limit.inc()
    self.assertRaises(cfg_utils.TooComplexError, limit.inc)

  def test_variable_product(self):
    u1 = self.prog.NewVariable([1, 2], [], self.current_location)
    u2 = self.prog.NewVariable([3, 4], [], self.current_location)
    product = cfg_utils.variable_product([u1, u2])
    pairs = [[a.data for a in d]
             for d in product]
    self.assertCountEqual(pairs, [
        [1, 3],
        [1, 4],
        [2, 3],
        [2, 4],
    ])

  def test_deep_variable_product_raises(self):
    x1, x2 = [DummyValue(i + 1) for i in range(2)]
    v1 = self.prog.NewVariable([x1, x2], [], self.current_location)
    v2 = self.prog.NewVariable([x1, x2], [], self.current_location)
    v3 = self.prog.NewVariable([x1, x2], [], self.current_location)
    v4 = self.prog.NewVariable([x1, x2], [], self.current_location)
    v5 = self.prog.NewVariable([x1, x2], [], self.current_location)
    v6 = self.prog.NewVariable([x1, x2], [], self.current_location)
    v7 = self.prog.NewVariable([x1, x2], [], self.current_location)
    v8 = self.prog.NewVariable([x1, x2], [], self.current_location)
    self.assertRaises(cfg_utils.TooComplexError,
                      cfg_utils.deep_variable_product,
                      [v1, v2, v3, v4, v5, v6, v7, v8],
                      256)

  def test_deep_variable_product_raises2(self):
    x1, x2, x3, x4 = [DummyValue(i + 1) for i in range(4)]
    v1 = self.prog.NewVariable([x1, x2], [], self.current_location)
    v2 = self.prog.NewVariable([x1, x2], [], self.current_location)
    v3 = self.prog.NewVariable([x3, x4], [], self.current_location)
    v4 = self.prog.NewVariable([x3, x4], [], self.current_location)
    x1.set_parameters([v3])
    x2.set_parameters([v4])
    self.assertRaises(cfg_utils.TooComplexError,
                      cfg_utils.deep_variable_product, [v1, v2], 4)

  def test_variable_product_dict_raises(self):
    values = [DummyValue(i + 1) for i in range(4)]
    v1 = self.prog.NewVariable(values, [], self.current_location)
    v2 = self.prog.NewVariable(values, [], self.current_location)
    v3 = self.prog.NewVariable(values, [], self.current_location)
    v4 = self.prog.NewVariable(values, [], self.current_location)
    variabledict = {"v1": v1, "v2": v2, "v3": v3, "v4": v4}
    self.assertRaises(cfg_utils.TooComplexError,
                      cfg_utils.variable_product_dict, variabledict, 4)

  def test_deep_variable_product(self):
    x1, x2, x3, x4, x5, x6 = [DummyValue(i + 1) for i in range(6)]
    v1 = self.prog.NewVariable([x1, x2], [], self.current_location)
    v2 = self.prog.NewVariable([x3], [], self.current_location)
    v3 = self.prog.NewVariable([x4, x5], [], self.current_location)
    v4 = self.prog.NewVariable([x6], [], self.current_location)
    x1.set_parameters([v2, v3])
    product = cfg_utils.deep_variable_product([v1, v4])
    rows = [{a.data for a in row}
            for row in product]
    self.assertCountEqual(rows, [
        {x1, x3, x4, x6},
        {x1, x3, x5, x6},
        {x2, x6},
    ])

  def test_deep_variable_product_with_empty_variables(self):
    x1 = DummyValue(1)
    v1 = self.prog.NewVariable([x1], [], self.current_location)
    v2 = self.prog.NewVariable([], [], self.current_location)
    x1.set_parameters([v2])
    product = cfg_utils.deep_variable_product([v1])
    rows = [{a.data for a in row}
            for row in product]
    self.assertCountEqual(rows, [{x1}])

  def test_deep_variable_product_with_empty_top_layer(self):
    x1 = DummyValue(1)
    v1 = self.prog.NewVariable([x1], [], self.current_location)
    v2 = self.prog.NewVariable([], [], self.current_location)
    product = cfg_utils.deep_variable_product([v1, v2])
    rows = [{a.data for a in row}
            for row in product]
    self.assertCountEqual(rows, [{x1}])

  def test_deep_variable_product_with_cycle(self):
    x1, x2, x3, x4, x5, x6 = [DummyValue(i + 1) for i in range(6)]
    v1 = self.prog.NewVariable([x1, x2], [], self.current_location)
    v2 = self.prog.NewVariable([x3], [], self.current_location)
    v3 = self.prog.NewVariable([x4, x5], [], self.current_location)
    v4 = self.prog.NewVariable([x6], [], self.current_location)
    x1.set_parameters([v2, v3])
    x5.set_parameters([v1])
    product = cfg_utils.deep_variable_product([v1, v4])
    rows = [{a.data for a in row}
            for row in product]
    self.assertCountEqual(rows, [
        {x1, x3, x4, x6},
        {x1, x2, x3, x5, x6},
        {x1, x3, x5, x6},
        {x2, x6},
    ])

  def test_variable_product_dict(self):
    u1 = self.prog.NewVariable([1, 2], [], self.current_location)
    u2 = self.prog.NewVariable([3, 4], [], self.current_location)
    product = cfg_utils.variable_product_dict({"a": u1, "b": u2})
    pairs = [{k: a.data for k, a in d.items()} for d in product]
    self.assertCountEqual(pairs, [
        {"a": 1, "b": 3},
        {"a": 1, "b": 4},
        {"a": 2, "b": 3},
        {"a": 2, "b": 4},
    ])


class Node:
  """A graph node, for testing topological sorting."""

  def __init__(self, name, *incoming):
    self.name = name
    self.outgoing = []
    self.incoming = list(incoming)
    for n in incoming:
      n.outgoing.append(self)

  def connect_to(self, other_node):
    self.outgoing.append(other_node)
    other_node.incoming.append(self)

  def __repr__(self):
    return "Node(%s)" % self.name


class GraphUtilTest(unittest.TestCase):
  """Test abstract graph utilities."""

  def setUp(self):
    super().setUp()
    self.prog = cfg.Program()

  def test_compute_predecessors(self):
    # n7      n6
    #  ^      ^
    #  |      |
    #  |      |
    # n1 ---> n20 --> n3 --> n5 -+
    #         | ^            ^   |
    #         | |            |   |
    #         | +------------|---+
    #         v              |
    #         n4 ------------+
    n1 = self.prog.NewCFGNode("n1")
    n20 = n1.ConnectNew("n20")
    n3 = n20.ConnectNew("n3")
    n4 = n20.ConnectNew("n4")
    n5 = n3.ConnectNew("n5")
    n6 = n20.ConnectNew("n6")
    n7 = n1.ConnectNew("n7")
    n3.ConnectTo(n5)
    n4.ConnectTo(n5)
    n5.ConnectTo(n20)

    # Intentionally pick a non-root as nodes[0] to verify that the graph
    # will still be fully explored.
    nodes = [n7, n1, n20, n3, n4, n5, n6]
    r = cfg_utils.compute_predecessors(nodes)
    self.assertCountEqual(r[n1], {n1})
    self.assertCountEqual(r[n20], {n1, n20, n3, n4, n5})
    self.assertCountEqual(r[n3], {n1, n20, n3, n4, n5})
    self.assertCountEqual(r[n4], {n1, n20, n3, n4, n5})
    self.assertCountEqual(r[n5], {n1, n20, n3, n4, n5})
    self.assertCountEqual(r[n6], {n1, n20, n3, n4, n5, n6})
    self.assertCountEqual(r[n7], {n1, n7})

  def test_order_nodes0(self):
    order = cfg_utils.order_nodes([])
    self.assertCountEqual(order, [])

  def test_order_nodes1(self):
    # n1 --> n2
    n1 = self.prog.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    order = cfg_utils.order_nodes([n1, n2])
    self.assertCountEqual([n1, n2], order)

  def test_order_nodes2(self):
    # n1   n2(dead)
    n1 = self.prog.NewCFGNode("n1")
    n2 = self.prog.NewCFGNode("n2")
    order = cfg_utils.order_nodes([n1, n2])
    self.assertCountEqual([n1], order)

  def test_order_nodes3(self):
    # n1 --> n2 --> n3
    # ^             |
    # +-------------+
    n1 = self.prog.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    n3 = n2.ConnectNew("n3")
    n3.ConnectTo(n1)
    order = cfg_utils.order_nodes([n1, n2, n3])
    self.assertCountEqual([n1, n2, n3], order)

  def test_order_nodes4(self):
    # n1 --> n3 --> n2
    # ^      |
    # +------+
    n1 = self.prog.NewCFGNode("n1")
    n3 = n1.ConnectNew("n3")
    n2 = n3.ConnectNew("n2")
    n3.ConnectTo(n1)
    order = cfg_utils.order_nodes([n1, n2, n3])
    self.assertCountEqual([n1, n3, n2], order)

  def test_order_nodes5(self):
    # n1 --> n3 --> n2
    # ^      |
    # +------+      n4(dead)
    n1 = self.prog.NewCFGNode("n1")
    n3 = n1.ConnectNew("n3")
    n2 = n3.ConnectNew("n2")
    n3.ConnectTo(n1)
    n4 = self.prog.NewCFGNode("n4")
    order = cfg_utils.order_nodes([n1, n2, n3, n4])
    self.assertCountEqual([n1, n3, n2], order)

  def test_order_nodes6(self):
    #  +-------------------+
    #  |                   v
    # n1 --> n2 --> n3 --> n5
    #        ^      |
    #        +------n4
    n1 = self.prog.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    n3 = n2.ConnectNew("n3")
    n4 = n3.ConnectNew("n4")
    n4.ConnectTo(n2)
    n5 = n3.ConnectNew("n5")
    n1.ConnectTo(n5)
    order = cfg_utils.order_nodes([n1, n5, n4, n3, n2])
    self.assertCountEqual([n1, n2, n3, n4, n5], order)

  def test_order_nodes7(self):
    #  +---------------------------------+
    #  |                                 v
    # n1 --> n2 --> n3 --> n4 --> n5 --> n6
    #        ^      |      ^      |
    #        |      v      |      v
    #        +------n7     +------n8
    n1 = self.prog.NewCFGNode("n1")
    n2 = n1.ConnectNew("n2")
    n3 = n2.ConnectNew("n3")
    n4 = n3.ConnectNew("n4")
    n5 = n4.ConnectNew("n5")
    n6 = n5.ConnectNew("n6")
    n7 = n3.ConnectNew("n7")
    n7.ConnectTo(n2)
    n8 = n5.ConnectNew("n8")
    n8.ConnectTo(n4)
    n1.ConnectTo(n6)
    order = cfg_utils.order_nodes([n1, n2, n3, n4, n5, n6, n7, n8])
    self.assertCountEqual([n1, n2, n3, n7, n4, n5, n8, n6], order)

  def test_topological_sort(self):
    n1 = Node("1")
    n2 = Node("2", n1)
    n3 = Node("3", n2)
    n4 = Node("4", n2, n3)
    for permutation in itertools.permutations([n1, n2, n3, n4]):
      self.assertEqual(list(cfg_utils.topological_sort(permutation)),
                       [n1, n2, n3, n4])

  def test_topological_sort2(self):
    n1 = Node("1")
    n2 = Node("2", n1)
    self.assertEqual(list(cfg_utils.topological_sort([n1, n2, 3, 4]))[-1], n2)

  def test_topological_sort_cycle(self):
    n1 = Node("1")
    n2 = Node("2")
    n1.incoming = [n2]
    n2.incoming = [n1]
    generator = cfg_utils.topological_sort([n1, n2])
    self.assertRaises(ValueError, list, generator)

  def test_topological_sort_sub_cycle(self):
    n1 = Node("1")
    n2 = Node("2")
    n3 = Node("3")
    n1.incoming = [n2]
    n2.incoming = [n1]
    n3.incoming = [n1, n2]
    generator = cfg_utils.topological_sort([n1, n2, n3])
    self.assertRaises(ValueError, list, generator)

  def test_topological_sort_getattr(self):
    self.assertEqual(list(cfg_utils.topological_sort([1])), [1])


if __name__ == "__main__":
  unittest.main()



from typing import Any, List

class Rule:
    id: Any
    prefix: Any
    status: Any
    expiration: Any
    transition: Any
    def __init__(
        self,
        id: Any | None = ...,
        prefix: Any | None = ...,
        status: Any | None = ...,
        expiration: Any | None = ...,
        transition: Any | None = ...,
    ) -> None: ...
    def startElement(self, name, attrs, connection): ...
    def endElement(self, name, value, connection): ...
    def to_xml(self): ...

class Expiration:
    days: Any
    date: Any
    def __init__(self, days: Any | None = ..., date: Any | None = ...) -> None: ...
    def startElement(self, name, attrs, connection): ...
    def endElement(self, name, value, connection): ...
    def to_xml(self): ...

class Transition:
    days: Any
    date: Any
    storage_class: Any
    def __init__(self, days: Any | None = ..., date: Any | None = ..., storage_class: Any | None = ...) -> None: ...
    def to_xml(self): ...

class Transitions(List[Transition]):
    transition_properties: int
    current_transition_property: int
    temp_days: Any
    temp_date: Any
    temp_storage_class: Any
    def __init__(self) -> None: ...
    def startElement(self, name, attrs, connection): ...
    def endElement(self, name, value, connection): ...
    def to_xml(self): ...
    def add_transition(self, days: Any | None = ..., date: Any | None = ..., storage_class: Any | None = ...): ...
    @property
    def days(self): ...
    @property
    def date(self): ...
    @property
    def storage_class(self): ...

class Lifecycle(List[Rule]):
    def startElement(self, name, attrs, connection): ...
    def endElement(self, name, value, connection): ...
    def to_xml(self): ...
    def add_rule(
        self,
        id: Any | None = ...,
        prefix: str = ...,
        status: str = ...,
        expiration: Any | None = ...,
        transition: Any | None = ...,
    ): ...



from typing import Any

from stripe.stripe_object import StripeObject as StripeObject

class ErrorObject(StripeObject):
    def refresh_from(
        self,
        values,
        api_key: Any | None = ...,
        partial: bool = ...,
        stripe_version: Any | None = ...,
        stripe_account: Any | None = ...,
        last_response: Any | None = ...,
    ): ...

class OAuthErrorObject(StripeObject):
    def refresh_from(
        self,
        values,
        api_key: Any | None = ...,
        partial: bool = ...,
        stripe_version: Any | None = ...,
        stripe_account: Any | None = ...,
        last_response: Any | None = ...,
    ): ...



"""Test errors.py."""

import collections
import csv
import os
import textwrap

from pytype import errors
from pytype import file_utils
from pytype import state as frame_state
from pytype.tests import test_utils

import unittest

_TEST_ERROR = "test-error"
_MESSAGE = "an error message on 'here'"


class ErrorTest(unittest.TestCase):

  @errors._error_name(_TEST_ERROR)
  def test_init(self):
    e = errors.Error(errors.SEVERITY_ERROR, _MESSAGE, filename="foo.py",
                     lineno=123, methodname="foo", keyword="here")
    self.assertEqual(errors.SEVERITY_ERROR, e._severity)
    self.assertEqual(_MESSAGE, e._message)
    self.assertEqual(e._name, _TEST_ERROR)
    self.assertEqual("foo.py", e._filename)
    self.assertEqual(123, e._lineno)
    self.assertEqual("foo", e._methodname)
    self.assertEqual("here", e.keyword)

  @errors._error_name(_TEST_ERROR)
  def test_with_stack(self):
    # Opcode of None.
    e = errors.Error.with_stack(None, errors.SEVERITY_ERROR, _MESSAGE,
                                keyword="here")
    self.assertEqual(errors.SEVERITY_ERROR, e._severity)
    self.assertEqual(_MESSAGE, e._message)
    self.assertEqual(e._name, _TEST_ERROR)
    self.assertEqual(None, e._filename)
    self.assertEqual(0, e._lineno)
    self.assertEqual(None, e._methodname)
    self.assertEqual("here", e.keyword)
    # Opcode of None.
    op = test_utils.FakeOpcode("foo.py", 123, "foo")
    e = errors.Error.with_stack(op.to_stack(), errors.SEVERITY_ERROR, _MESSAGE,
                                keyword="here")
    self.assertEqual(errors.SEVERITY_ERROR, e._severity)
    self.assertEqual(_MESSAGE, e._message)
    self.assertEqual(e._name, _TEST_ERROR)
    self.assertEqual("foo.py", e._filename)
    self.assertEqual(123, e._lineno)
    self.assertEqual("foo", e._methodname)
    self.assertEqual("here", e.keyword)

  @errors._error_name(_TEST_ERROR)
  def test_no_traceback_stack_len_1(self):
    # Stack of length 1
    op = test_utils.FakeOpcode("foo.py", 123, "foo")
    error = errors.Error.with_stack(op.to_stack(), errors.SEVERITY_ERROR, "")
    self.assertIsNone(error._traceback)

  @errors._error_name(_TEST_ERROR)
  def test_no_traceback_no_opcode(self):
    # Frame without opcode
    op = test_utils.FakeOpcode("foo.py", 123, "foo")
    stack = [frame_state.SimpleFrame(), frame_state.SimpleFrame(op)]
    error = errors.Error.with_stack(stack, errors.SEVERITY_ERROR, "")
    self.assertIsNone(error._traceback)

  @errors._error_name(_TEST_ERROR)
  def test_traceback(self):
    stack = test_utils.fake_stack(errors.MAX_TRACEBACK_LENGTH + 1)
    error = errors.Error.with_stack(stack, errors.SEVERITY_ERROR, "")
    self.assertMultiLineEqual(error._traceback, textwrap.dedent("""
      Called from (traceback):
        line 0, in function0
        line 1, in function1
        line 2, in function2""").lstrip())

  @errors._error_name(_TEST_ERROR)
  def test_truncated_traceback(self):
    stack = test_utils.fake_stack(errors.MAX_TRACEBACK_LENGTH + 2)
    error = errors.Error.with_stack(stack, errors.SEVERITY_ERROR, "")
    self.assertMultiLineEqual(error._traceback, textwrap.dedent("""
      Called from (traceback):
        line 0, in function0
        ...
        line 3, in function3""").lstrip())

  def test__error_name(self):
    # This should be true as long as at least one method is annotated with
    # _error_name(_TEST_ERROR).
    self.assertIn(_TEST_ERROR, errors._ERROR_NAMES)

  def test_no_error_name(self):
    # It is illegal to create an error outside of an @error_name annotation.
    self.assertRaises(AssertionError, errors.Error, errors.SEVERITY_ERROR,
                      _MESSAGE)

  @errors._error_name(_TEST_ERROR)
  def test_str(self):
    e = errors.Error(errors.SEVERITY_ERROR, _MESSAGE, filename="foo.py",
                     lineno=123, methodname="foo")
    message = "an error message on 'here'"
    error = "File \"foo.py\", line 123, in foo: " + message + " [test-error]"
    self.assertEqual(error, str(e))

  @errors._error_name(_TEST_ERROR)
  def test_write_to_csv(self):
    errorlog = errors.ErrorLog()
    op = test_utils.FakeOpcode("foo.py", 123, "foo")
    message, details = "This is an error", "with\nsome\ndetails: \"1\", 2, 3"
    errorlog.error(op.to_stack(), message, details + "0")
    errorlog.error(op.to_stack(), message, details + "1")
    with file_utils.Tempdir() as d:
      filename = d.create_file("errors.csv")
      errorlog.print_to_csv_file(filename)
      with open(filename, "r") as fi:
        rows = list(csv.reader(fi, delimiter=","))
        self.assertEqual(len(rows), 2)
        for i, row in enumerate(rows):
          filename, lineno, name, actual_message, actual_details = row
          self.assertEqual(filename, "foo.py")
          self.assertEqual(lineno, "123")
          self.assertEqual(name, _TEST_ERROR)
          self.assertEqual(actual_message, message)
          self.assertEqual(actual_details, details + str(i))

  @errors._error_name(_TEST_ERROR)
  def test_write_to_csv_with_traceback(self):
    errorlog = errors.ErrorLog()
    stack = test_utils.fake_stack(2)
    errorlog.error(stack, "", "some\ndetails")
    with file_utils.Tempdir() as d:
      filename = d.create_file("errors.csv")
      errorlog.print_to_csv_file(filename)
      with open(filename, "r") as fi:
        (_, _, _, _, actual_details), = list(csv.reader(fi, delimiter=","))
        self.assertMultiLineEqual(actual_details, textwrap.dedent("""
          some
          details

          Called from (traceback):
            line 0, in function0""").lstrip())


class ErrorLogBaseTest(unittest.TestCase):

  @errors._error_name(_TEST_ERROR)
  def test_error(self):
    errorlog = errors.ErrorLog()
    op = test_utils.FakeOpcode("foo.py", 123, "foo")
    errorlog.error(op.to_stack(), "unknown attribute %s" % "xyz")
    self.assertEqual(len(errorlog), 1)
    e = list(errorlog)[0]  # iterate the log and save the first error.
    self.assertEqual(errors.SEVERITY_ERROR, e._severity)
    self.assertEqual("unknown attribute xyz", e._message)
    self.assertEqual(e._name, _TEST_ERROR)
    self.assertEqual("foo.py", e._filename)

  @errors._error_name(_TEST_ERROR)
  def test_error_with_details(self):
    errorlog = errors.ErrorLog()
    errorlog.error(None, "My message", "one\ntwo")
    self.assertEqual(textwrap.dedent("""
        My message [test-error]
          one
          two
        """).lstrip(), str(errorlog))

  @errors._error_name(_TEST_ERROR)
  def test_warn(self):
    errorlog = errors.ErrorLog()
    op = test_utils.FakeOpcode("foo.py", 123, "foo")
    errorlog.warn(op.to_stack(), "unknown attribute %s", "xyz")
    self.assertEqual(len(errorlog), 1)
    e = list(errorlog)[0]  # iterate the log and save the first error.
    self.assertEqual(errors.SEVERITY_WARNING, e._severity)
    self.assertEqual("unknown attribute xyz", e._message)
    self.assertEqual(e._name, _TEST_ERROR)
    self.assertEqual("foo.py", e._filename)

  @errors._error_name(_TEST_ERROR)
  def test_has_error(self):
    errorlog = errors.ErrorLog()
    self.assertFalse(errorlog.has_error())
    # A warning is part of the error log, but isn't severe.
    errorlog.warn(None, "A warning")
    self.assertEqual(len(errorlog), 1)
    self.assertFalse(errorlog.has_error())
    # An error is severe.
    errorlog.error(None, "An error")
    self.assertEqual(len(errorlog), 2)
    self.assertTrue(errorlog.has_error())

  @errors._error_name(_TEST_ERROR)
  def test_duplicate_error_no_traceback(self):
    errorlog = errors.ErrorLog()
    stack = test_utils.fake_stack(2)
    errorlog.error(stack, "error")  # traceback
    errorlog.error(stack[-1:], "error")  # no traceback
    # Keep the error with no traceback.
    unique_errors = errorlog.unique_sorted_errors()
    self.assertEqual(len(unique_errors), 1)
    self.assertIsNone(unique_errors[0]._traceback)

  @errors._error_name(_TEST_ERROR)
  def test_error_without_stack(self):
    errorlog = errors.ErrorLog()
    stack = test_utils.fake_stack(1)
    errorlog.error(stack, "error_with_stack")
    errorlog.error([], "error_without_stack")
    unique_errors = errorlog.unique_sorted_errors()
    unique_errors = [(error.message, error.filename, error.lineno)
                     for error in unique_errors
                    ]
    self.assertEqual(
        [("error_without_stack", None, 0),
         ("error_with_stack", "foo.py", 0)],
        unique_errors,
    )

  @errors._error_name(_TEST_ERROR)
  def test_duplicate_error_shorter_traceback(self):
    errorlog = errors.ErrorLog()
    stack = test_utils.fake_stack(3)
    errorlog.error(stack, "error")  # longer traceback
    errorlog.error(stack[-2:], "error")  # shorter traceback
    # Keep the error with a shorter traceback.
    unique_errors = errorlog.unique_sorted_errors()
    self.assertEqual(len(unique_errors), 1)
    self.assertMultiLineEqual(unique_errors[0]._traceback, textwrap.dedent("""
      Called from (traceback):
        line 1, in function1""").lstrip())

  @errors._error_name(_TEST_ERROR)
  def test_unique_errors(self):
    errorlog = errors.ErrorLog()
    current_frame = frame_state.SimpleFrame(
        test_utils.FakeOpcode("foo.py", 123, "foo"))
    backframe1 = frame_state.SimpleFrame(
        test_utils.FakeOpcode("foo.py", 1, "bar"))
    backframe2 = frame_state.SimpleFrame(
        test_utils.FakeOpcode("foo.py", 2, "baz"))
    errorlog.error([backframe1, current_frame], "error")
    errorlog.error([backframe2, current_frame], "error")
    # Keep both errors, since the tracebacks are different.
    unique_errors = errorlog.unique_sorted_errors()
    self.assertEqual(len(unique_errors), 2)
    self.assertSetEqual(set(errorlog), set(unique_errors))


class ErrorDocTest(unittest.TestCase):
  ERROR_FILE_PATH = os.path.join(os.path.dirname(errors.__file__),
                                 "../docs/errors.md")

  def _check_and_get_documented_errors(self):
    with open(self.ERROR_FILE_PATH, "r") as f:
      lines = f.readlines()
    entries = [line[3:].strip() for line in lines if line.startswith("##")]
    counts = collections.Counter(entries)
    if any(count > 1 for count in counts.values()):
      raise AssertionError(
          "These errors.md entries have duplicates:\n  " +
          "\n  ".join(name for name, count in counts.items() if count > 1))
    sorted_entries = sorted(entries)
    if sorted_entries != entries:
      raise AssertionError(
          "These errors.md entries are out of alphabetical order:\n  " +
          "\n  ".join(
              actual for expected, actual in zip(sorted_entries, entries)
              if expected != actual))
    return set(entries)

  def test_error_doc(self):
    documented_errors = self._check_and_get_documented_errors()
    all_errors = errors._ERROR_NAMES - {_TEST_ERROR}
    undocumented_errors = all_errors - documented_errors
    if undocumented_errors:
      raise AssertionError(
          "These errors are missing entries in errors.md:\n  " +
          "\n  ".join(undocumented_errors) + "\n")
    deprecated_errors = documented_errors - all_errors
    if deprecated_errors:
      raise AssertionError(
          "These errors.md entries do not correspond to errors:\n  " +
          "\n  ".join(deprecated_errors))


if __name__ == "__main__":
  unittest.main()



from typing import Any, MutableMapping

class URIDict(MutableMapping[Any, Any]):
    def normalize(self, uri): ...
    store: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def __getitem__(self, uri): ...
    def __setitem__(self, uri, value) -> None: ...
    def __delitem__(self, uri) -> None: ...
    def __iter__(self): ...
    def __len__(self): ...

class Unset: ...

def load_schema(name): ...
def indent(string, times: int = ...): ...
def format_as_index(indices): ...
def find_additional_properties(instance, schema) -> None: ...
def extras_msg(extras): ...
def types_msg(instance, types): ...
def flatten(suitable_for_isinstance): ...
def ensure_list(thing): ...
def equal(one, two): ...
def unbool(element, true=..., false=...): ...
def uniq(container): ...



"""Implementation of special members of typing_extensions."""
from pytype import overlay
from pytype.overlays import typing_overlay


class TypingExtensionsOverlay(overlay.Overlay):
  """A custom overlay for the 'typing_extensions' module."""

  def __init__(self, ctx):
    ast = ctx.loader.import_name("typing_extensions")
    member_map = {
        "Annotated": typing_overlay.typing_overlay["Annotated"],
        "final": typing_overlay.typing_overlay["final"],
        "Literal": typing_overlay.typing_overlay["Literal"],
        "Protocol": _build("typing.Protocol"),
        "runtime": _build("typing.runtime_checkable"),
        "SupportsIndex": _build("typing_extensions.SupportsIndex", ast),
        "TypedDict": typing_overlay.typing_overlay["TypedDict"],
    }
    for pyval in ast.aliases + ast.classes + ast.constants + ast.functions:
      # Any public typing_extensions members that are not explicitly implemented
      # are unsupported.
      _, name = pyval.name.rsplit(".", 1)
      if name.startswith("_"):
        continue
      if f"typing.{name}" not in ctx.loader.typing:
        if name not in member_map:
          member_map[name] = overlay.build(
              name, typing_overlay.not_supported_yet)
    super().__init__(ctx, "typing_extensions", member_map, ast)

  def _convert_member(self, member, subst=None):
    var = super()._convert_member(member, subst)
    for val in var.data:
      # typing_extensions backports typing features to older versions.
      # Pretending that the backports are in typing is easier than remembering
      # to check for both typing.X and typing_extensions.X every time we match
      # on an abstract value.
      val.module = "typing"
    return var


def _build(name, ast=None):
  return lambda ctx: ctx.convert.name_to_value(name, ast=ast)



from typing import Any

from stripe import oauth as oauth
from stripe.api_resources.abstract import (
    CreateableAPIResource as CreateableAPIResource,
    DeletableAPIResource as DeletableAPIResource,
    ListableAPIResource as ListableAPIResource,
    UpdateableAPIResource as UpdateableAPIResource,
    custom_method as custom_method,
    nested_resource_class_methods as nested_resource_class_methods,
)

class Account(CreateableAPIResource, DeletableAPIResource, ListableAPIResource):
    OBJECT_NAME: str
    def reject(self, idempotency_key: Any | None = ..., **params): ...
    @classmethod
    def retrieve(cls, id: Any | None = ..., api_key: Any | None = ..., **params): ...
    @classmethod
    def modify(cls, id: Any | None = ..., **params): ...
    def instance_url(self): ...
    def persons(self, **params): ...
    def deauthorize(self, **params): ...
    def serialize(self, previous): ...



from typing import Any

from stripe.api_resources.abstract import UpdateableAPIResource as UpdateableAPIResource
from stripe.api_resources.transfer import Transfer as Transfer

class Reversal(UpdateableAPIResource):
    OBJECT_NAME: str
    def instance_url(self): ...
    @classmethod
    def modify(cls, sid, **params) -> None: ...
    @classmethod
    def retrieve(cls, id, api_key: Any | None = ..., **params) -> None: ...



"""Internal representation of pytd nodes.

All pytd nodes should be frozen attrs inheriting from node.Node (aliased
to Node in this module). Nodes representing types should also inherit from Type.

Since we use frozen classes, setting attributes in __post_init__ needs to be
done via
  object.__setattr__(self, 'attr', value)

NOTE: The way we introspect on the types of the fields requires forward
references to be simple classes, hence use
  x: Union['Foo', 'Bar']
rather than
  x: 'Union[Foo, Bar]'
"""

import collections
import itertools

import typing
from typing import Any, Optional, Tuple, TypeVar, Union

import attr

from pytype.pytd.parse import node

# Alias node.Node for convenience.
Node = node.Node

_TypeT = TypeVar('_TypeT', bound='Type')


class Type:
  """Each type class below should inherit from this mixin."""
  name: Optional[str]

  # We type-annotate many things as pytd.Type when we'd really want them to be
  # Intersection[pytd.Type, pytd.parse.node.Node], so we need to copy the type
  # signature of Node.Visit here.
  if typing.TYPE_CHECKING:

    def Visit(self: _TypeT, visitor, *args, **kwargs) -> _TypeT:
      del visitor, args, kwargs  # unused
      return self

  __slots__ = ()


@attr.s(auto_attribs=True, frozen=True, order=False, eq=False)
class TypeDeclUnit(Node):
  """Module node. Holds module contents (constants / classes / functions).

  Attributes:
    name: Name of this module, or None for the top-level module.
    constants: Iterable of module-level constants.
    type_params: Iterable of module-level type parameters.
    functions: Iterable of functions defined in this type decl unit.
    classes: Iterable of classes defined in this type decl unit.
    aliases: Iterable of aliases (or imports) for types in other modules.
  """
  name: Optional[str]
  constants: Tuple['Constant', ...]
  type_params: Tuple['TypeParameter', ...]
  classes: Tuple['Class', ...]
  functions: Tuple['Function', ...]
  aliases: Tuple['Alias', ...]

  def Lookup(self, name):
    """Convenience function: Look up a given name in the global namespace.

    Tries to find a constant, function or class by this name.

    Args:
      name: Name to look up.

    Returns:
      A Constant, Function or Class.

    Raises:
      KeyError: if this identifier doesn't exist.
    """
    # TODO(b/159053187): Put constants, functions, classes and aliases into a
    # combined dict.
    try:
      return self._name2item[name]
    except AttributeError:
      object.__setattr__(self, '_name2item', {})
      for x in self.type_params:
        self._name2item[x.full_name] = x
      for x in self.constants + self.functions + self.classes + self.aliases:
        self._name2item[x.name] = x
      return self._name2item[name]

  def __contains__(self, name):
    try:
      self.Lookup(name)
    except KeyError:
      return False
    return True

  # The hash/eq/ne values are used for caching and speed things up quite a bit.

  def __hash__(self):
    return id(self)

  def __eq__(self, other):
    return id(self) == id(other)

  def __ne__(self, other):
    return id(self) != id(other)


@attr.s(auto_attribs=True, frozen=True, order=False, slots=True,
        cache_hash=True)
class Constant(Node):
  name: str
  type: Type
  value: Any = None


@attr.s(auto_attribs=True, frozen=True, order=False, slots=True,
        cache_hash=True)
class Alias(Node):
  """An alias (symbolic link) for a class implemented in some other module.

  Unlike Constant, the Alias is the same type, as opposed to an instance of that
  type. It's generated, among others, from imports - e.g. "from x import y as z"
  will create a local alias "z" for "x.y".
  """
  name: str
  type: Union[Type, Constant, 'Function', 'Module']


@attr.s(auto_attribs=True, frozen=True, order=False, slots=True,
        cache_hash=True)
class Module(Node):
  """A module imported into the current module, possibly with an alias."""
  name: str
  module_name: str

  @property
  def is_aliased(self):
    return self.name != self.module_name


@attr.s(auto_attribs=True, frozen=True, order=False, cache_hash=True)
class Class(Node):
  """Represents a class declaration.

  Used as dict/set key, so all components must be hashable.

  Attributes:
    name: Class name (string)
    bases: The super classes of this class (instances of pytd.Type).
    methods: Tuple of methods, classmethods, staticmethods
      (instances of pytd.Function).
    constants: Tuple of constant class attributes (instances of pytd.Constant).
    classes: Tuple of nested classes.
    slots: A.k.a. __slots__, declaring which instance attributes are writable.
    template: Tuple of pytd.TemplateItem instances.
  """
  name: str
  metaclass: Union[None, Type]
  bases: Tuple[Union['Class', Type], ...]
  methods: Tuple['Function', ...]
  constants: Tuple[Constant, ...]
  classes: Tuple['Class', ...]
  decorators: Tuple[Alias, ...]
  slots: Optional[Tuple[str, ...]]
  template: Tuple['TemplateItem', ...]

  def Lookup(self, name):
    """Convenience function: Look up a given name in the class namespace.

    Tries to find a method or constant by this name in the class.

    Args:
      name: Name to look up.

    Returns:
      A Constant or Function instance.

    Raises:
      KeyError: if this identifier doesn't exist in this class.
    """
    # TODO(b/159053187): Remove this. Make methods and constants dictionaries.
    try:
      return self._name2item[name]
    except AttributeError:
      object.__setattr__(self, '_name2item', {})
      for x in self.methods + self.constants + self.classes:
        self._name2item[x.name] = x
      return self._name2item[name]


class MethodTypes:
  METHOD = 'method'
  STATICMETHOD = 'staticmethod'
  CLASSMETHOD = 'classmethod'
  PROPERTY = 'property'


class MethodFlags:
  ABSTRACT = 1
  COROUTINE = 2

  @classmethod
  def abstract_flag(cls, is_abstract):  # pylint: disable=invalid-name
    # Useful when creating functions directly (other flags aren't needed there).
    return cls.ABSTRACT if is_abstract else 0


@attr.s(auto_attribs=True, frozen=True, order=False, slots=True,
        cache_hash=True)
class Function(Node):
  """A function or a method, defined by one or more PyTD signatures.

  Attributes:
    name: The name of this function.
    signatures: Tuple of possible parameter type combinations for this function.
    kind: The type of this function. One of: STATICMETHOD, CLASSMETHOD, METHOD
    flags: A bitfield of flags like is_abstract
  """
  name: str
  signatures: Tuple['Signature', ...]
  kind: str
  flags: int = 0

  @property
  def is_abstract(self):
    return bool(self.flags & MethodFlags.ABSTRACT)

  @property
  def is_coroutine(self):
    return bool(self.flags & MethodFlags.COROUTINE)

  def with_flag(self, flag, value):
    """Return a copy of self with flag set to value."""
    new_flags = self.flags | flag if value else self.flags & ~flag
    return self.Replace(flags=new_flags)


@attr.s(auto_attribs=True, frozen=True, order=False, slots=True,
        cache_hash=True)
class Signature(Node):
  """Represents an individual signature of a function.

  For overloaded functions, this is one specific combination of parameters.
  For non-overloaded functions, there is a 1:1 correspondence between function
  and signature.

  Attributes:
    params: The list of parameters for this function definition.
    starargs: Name of the "*" parameter. The "args" in "*args".
    starstarargs: Name of the "*" parameter. The "kw" in "**kw".
    return_type: The return type of this function.
    exceptions: List of exceptions for this function definition.
    template: names for bindings for bounded types in params/return_type
  """
  params: Tuple['Parameter', ...]
  starargs: Optional['Parameter']
  starstarargs: Optional['Parameter']
  return_type: Type
  exceptions: Tuple[Type, ...]
  template: Tuple['TemplateItem', ...]

  @property
  def name(self):
    return None

  @property
  def has_optional(self):
    return self.starargs is not None or self.starstarargs is not None


@attr.s(auto_attribs=True, frozen=True, order=False, slots=True,
        cache_hash=True)
class Parameter(Node):
  """Represents a parameter of a function definition.

  Attributes:
    name: The name of the parameter.
    type: The type of the parameter.
    kwonly: True if this parameter can only be passed as a keyword parameter.
    optional: If the parameter is optional.
    mutated_type: The type the parameter will have after the function is called
      if the type is mutated, None otherwise.
  """
  name: str
  type: Type
  kwonly: bool
  optional: bool
  mutated_type: Optional[Type]


@attr.s(auto_attribs=True, frozen=True, order=False, slots=True,
        cache_hash=True)
class TypeParameter(Node, Type):
  """Represents a type parameter.

  A type parameter is a bound variable in the context of a function or class
  definition. It specifies an equivalence between types.
  For example, this defines an identity function:
    def f(x: T) -> T

  Attributes:
    name: Name of the parameter. E.g. "T".
    scope: Fully-qualified name of the class or function this parameter is
      bound to. E.g. "mymodule.MyClass.mymethod", or None.
  """
  name: str
  constraints: Tuple[Type, ...] = attr.ib(factory=tuple)
  bound: Optional[Type] = None
  scope: Optional[str] = None

  def __lt__(self, other):
    try:
      return super().__lt__(other)
    except TypeError:
      # In Python 3, str and None are not comparable. Declare None to be less
      # than every str so that visitors.AdjustTypeParameters.VisitTypeDeclUnit
      # can sort type parameters.
      return self.scope is None

  @property
  def full_name(self):
    # There are hard-coded type parameters in the code (e.g., T for sequences),
    # so the full name cannot be stored directly in self.name. Note that "" is
    # allowed as a module name.
    return ('' if self.scope is None else self.scope + '.') + self.name

  @property
  def upper_value(self):
    if self.constraints:
      return UnionType(self.constraints)
    elif self.bound:
      return self.bound
    else:
      return AnythingType()


@attr.s(auto_attribs=True, frozen=True, order=False, slots=True,
        cache_hash=True)
class TemplateItem(Node):
  """Represents template name for generic types.

  This is used for classes and signatures. The 'template' field of both is
  a list of TemplateItems. Note that *using* the template happens through
  TypeParameters.  E.g. in:
    class A<T>:
      def f(T x) -> T
  both the "T"s in the definition of f() are using pytd.TypeParameter to refer
  to the TemplateItem in class A's template.

  Attributes:
    type_param: the TypeParameter instance used. This is the actual instance
      that's used wherever this type parameter appears, e.g. within a class.
  """
  type_param: TypeParameter

  @property
  def name(self):
    return self.type_param.name

  @property
  def full_name(self):
    return self.type_param.full_name


# Types can be:
# 1.) NamedType:
#     Specifies a type or import by name.
# 2.) ClassType
#     Points back to a Class in the AST. (This makes the AST circular)
# 3.) GenericType
#     Contains a base type and parameters.
# 4.) UnionType
#     Can be multiple types at once.
# 5.) NothingType / AnythingType
#     Special purpose types that represent nothing or everything.
# 6.) TypeParameter
#     A placeholder for a type.
# For 1 and 2, the file visitors.py contains tools for converting between the
# corresponding AST representations.


@attr.s(auto_attribs=True, frozen=True, order=False, slots=True,
        cache_hash=True)
class NamedType(Node, Type):
  """A type specified by name and, optionally, the module it is in."""
  name: str

  def __str__(self):
    return self.name


@attr.s(auto_attribs=False, init=False, frozen=False, order=False, slots=False,
        eq=False)
class ClassType(Node, Type):
  """A type specified through an existing class node."""
  # This type is different from normal nodes:
  # (a) It's mutable, and there are functions
  #     (parse/visitors.py:FillInLocalPointers) that modify a tree in place.
  # (b) The cls pointer is not treated as a regular attr field.
  # (c) Visitors will not process the "children" of this node. Since we point
  #     to classes that are back at the top of the tree, that would generate
  #     cycles.

  name: str = attr.ib()

  # We do not want cls to be a child node, but we do want it to be an optional
  # arg to __init__ and accessible via self.cls
  cls = None

  def __init__(self, name, cls=None):
    self.name = name
    self.cls = cls

  def __eq__(self, other):
    return (self.__class__ == other.__class__ and
            self.name == other.name)

  def __hash__(self):
    return hash((self.__class__.__name__, self.name))

  def __str__(self):
    return str(self.cls.name) if self.cls else self.name

  def __repr__(self):
    return '{type}{cls}({name})'.format(
        type=type(self).__name__, name=self.name,
        cls='<unresolved>' if self.cls is None else '')


@attr.s(auto_attribs=True, frozen=True, order=False, slots=True,
        cache_hash=True)
class LateType(Node, Type):
  """A type we have yet to resolve."""
  name: str
  recursive: bool = False

  def __str__(self):
    return self.name


@attr.s(auto_attribs=True, frozen=True, order=False, slots=True,
        cache_hash=True)
class AnythingType(Node, Type):
  """A type we know nothing about yet (? in pytd)."""

  @property
  def name(self):
    return None

  def __bool__(self):
    return True


@attr.s(auto_attribs=True, frozen=True, order=False, slots=True,
        cache_hash=True)
class NothingType(Node, Type):
  """An "impossible" type, with no instances (nothing in pytd).

  Also known as the "uninhabited" type, or, in type systems, the "bottom" type.
  For representing empty lists, and functions that never return.
  """

  @property
  def name(self):
    return None

  def __bool__(self):
    return True


def _FlattenTypes(type_list) -> Tuple[Type, ...]:
  """Helper function for _SetOfTypes initialization."""
  assert type_list  # Disallow empty sets. Use NothingType for these.
  flattened = itertools.chain.from_iterable(
      t.type_list if isinstance(t, _SetOfTypes) else [t]
      for t in type_list)
  # Remove duplicates, preserving order
  unique = tuple(collections.OrderedDict.fromkeys(flattened))
  return unique


@attr.s(auto_attribs=True, frozen=True, order=False, slots=True, eq=False)
class _SetOfTypes(Node, Type):
  """Super class for shared behavior of UnionType and IntersectionType."""
  # NOTE: type_list is kept as a tuple, to preserve the original order
  #       even though in most respects it acts like a frozenset.
  #       It also flattens the input, such that printing without
  #       parentheses gives the same result.
  type_list: Tuple[Type, ...] = attr.ib(converter=_FlattenTypes)

  @property
  def name(self):
    return None

  def __hash__(self):
    # See __eq__ - order doesn't matter, so use frozenset
    return hash(frozenset(self.type_list))

  def __eq__(self, other):
    if self is other:
      return True
    if isinstance(other, type(self)):
      # equality doesn't care about the ordering of the type_list
      return frozenset(self.type_list) == frozenset(other.type_list)
    return NotImplemented

  def __ne__(self, other):
    return not self == other


class UnionType(_SetOfTypes):
  """A union type that contains all types in self.type_list."""
  __slots__ = ()


class IntersectionType(_SetOfTypes):
  """An intersection type."""
  __slots__ = ()


@attr.s(auto_attribs=True, frozen=True, order=False, slots=True,
        cache_hash=True)
class GenericType(Node, Type):
  """Generic type. Takes a base type and type parameters.

  This is used for homogeneous tuples, lists, dictionaries, user classes, etc.

  Attributes:
    base_type: The base type. Instance of Type.
    parameters: Type parameters. Tuple of instances of Type.
  """
  base_type: Union[NamedType, ClassType, LateType]
  parameters: Tuple[Type, ...]

  @property
  def name(self):
    return self.base_type.name

  @property
  def element_type(self):
    """Type of the contained type, assuming we only have one type parameter."""
    element_type, = self.parameters
    return element_type


class TupleType(GenericType):
  """Special generic type for heterogeneous tuples.

  A tuple with length len(self.parameters), whose item type is specified at
  each index.
  """
  __slots__ = ()


class CallableType(GenericType):
  """Special generic type for a Callable that specifies its argument types.

  A Callable with N arguments has N+1 parameters. The first N parameters are
  the individual argument types, in the order of the arguments, and the last
  parameter is the return type.
  """
  __slots__ = ()

  @property
  def args(self):
    return self.parameters[:-1]

  @property
  def ret(self):
    return self.parameters[-1]


@attr.s(auto_attribs=True, frozen=True, order=False, slots=True,
        cache_hash=True)
class Literal(Node, Type):
  value: Union[int, str, Type]

  @property
  def name(self):
    return None


@attr.s(auto_attribs=True, frozen=True, order=False, slots=True,
        cache_hash=True)
class Annotated(Node, Type):
  base_type: Type
  annotations: Tuple[str, ...]

  @property
  def name(self):
    return None


# Types that can be a base type of GenericType:
GENERIC_BASE_TYPE = (NamedType, ClassType)


def IsContainer(t):
  """Checks whether class t is a container."""
  assert isinstance(t, Class)
  if t.name in ('typing.Generic', 'typing.Protocol'):
    return True
  for p in t.bases:
    if isinstance(p, GenericType):
      base = p.base_type
      # We need to check for Generic and Protocol again here because base may
      # not yet have been resolved to a ClassType.
      if (base.name in ('typing.Generic', 'typing.Protocol') or
          isinstance(base, ClassType) and IsContainer(base.cls)):
        return True
  return False


# Singleton objects that will be automatically converted to their types.
# The unqualified form is there so local name resolution can special-case it.
SINGLETON_TYPES = frozenset({'Ellipsis', 'builtins.Ellipsis'})


def ToType(item, allow_constants=False, allow_functions=False,
           allow_singletons=False):
  """Convert a pytd AST item into a type.

  Takes an AST item representing the definition of a type and returns an item
  representing a reference to the type. For example, if the item is a
  pytd.Class, this method will return a pytd.ClassType whose cls attribute
  points to the class.

  Args:
    item: A pytd.Node item.
    allow_constants: When True, constants that cannot be converted to types will
      be passed through unchanged.
    allow_functions: When True, functions that cannot be converted to types will
      be passed through unchanged.
    allow_singletons: When True, singletons that act as their types in
      annotations will return that type.

  Returns:
    A pytd.Type object representing the type of an instance of `item`.
  """
  if isinstance(item, Type):
    return item
  elif isinstance(item, Module):
    return item
  elif isinstance(item, Class):
    return ClassType(item.name, item)
  elif isinstance(item, Function) and allow_functions:
    return item
  elif isinstance(item, Constant):
    if allow_singletons and item.name in SINGLETON_TYPES:
      return item.type
    elif item.type.name == 'builtins.type':
      # A constant whose type is Type[C] is equivalent to class C, so the type
      # of an instance of the constant is C.
      if isinstance(item.type, GenericType):
        return item.type.parameters[0]
      else:
        return AnythingType()
    elif item.type.name == 'typing_extensions._SpecialForm':
      # We convert known special forms to their corresponding types and
      # otherwise treat them as unknown types.
      if item.name == 'typing_extensions.Protocol':
        return NamedType('typing.Protocol')
      else:
        return AnythingType()
    elif (isinstance(item.type, AnythingType) or
          item.name == 'typing_extensions.TypedDict'):
      # A constant with type Any may be a type, and TypedDict is a class that
      # looks like a constant:
      #   https://github.com/python/typeshed/blob/8cad322a8ccf4b104cafbac2c798413edaa4f327/third_party/2and3/typing_extensions.pyi#L68
      return AnythingType()
    elif allow_constants:
      return item
  elif isinstance(item, Alias):
    return item.type
  raise NotImplementedError("Can't convert %s: %s" % (type(item), item))


def AliasMethod(func, from_constant):
  """Returns method func with its signature modified as if it has been aliased.

  Args:
    func: A pytd.Function.
    from_constant: If True, func will be modified as if it has been aliased from
      an instance of its defining class, e.g.,
        class Foo:
          def func(self): ...
        const = ...  # type: Foo
        func = const.func
      Otherwise, it will be modified as if aliased from the class itself:
        class Foo:
          def func(self): ...
        func = Foo.func

  Returns:
    A pytd.Function, the aliased method.
  """
  # We allow module-level aliases of methods from classes and class instances.
  # When a static method is aliased, or a normal method is aliased from a class
  # (not an instance), the entire method signature is copied. Otherwise, the
  # first parameter ('self' or 'cls') is dropped.
  new_func = func.Replace(kind=MethodTypes.METHOD)
  if func.kind == MethodTypes.STATICMETHOD or (
      func.kind == MethodTypes.METHOD and not from_constant):
    return new_func
  return new_func.Replace(signatures=tuple(
      s.Replace(params=s.params[1:]) for s in new_func.signatures))


def LookupItemRecursive(module, name):
  """Recursively look up name in module."""
  parts = name.split('.')
  partial_name = module.name
  prev_item = None
  item = module

  def ExtractClass(t):
    if isinstance(t, ClassType) and t.cls:
      return t.cls
    t = module.Lookup(t.name)  # may raise KeyError
    if isinstance(t, Class):
      return t
    raise KeyError(t.name)

  for part in parts:
    prev_item = item
    # Check the type of item and give up if we encounter a type we don't know
    # how to handle.
    if isinstance(item, Constant):
      item = ExtractClass(item.type)  # may raise KeyError
    elif not isinstance(item, (TypeDeclUnit, Class)):
      raise KeyError(name)
    lookup_name = partial_name + '.' + part

    def Lookup(item, *names):
      for name in names:
        try:
          return item.Lookup(name)
        except KeyError:
          continue
      raise KeyError(names[-1])

    # Nested class names are fully qualified while function names are not, so
    # we try lookup for both naming conventions.
    try:
      item = Lookup(item, lookup_name, part)
    except KeyError:
      if not isinstance(item, Class):
        raise
      for base in item.bases:
        base_cls = ExtractClass(base)  # may raise KeyError
        try:
          item = Lookup(base_cls, lookup_name, part)
        except KeyError:
          continue  # continue up the MRO
        else:
          break  # name found!
      else:
        raise  # unresolved
    if isinstance(item, Constant):
      partial_name += '.' + item.name.rsplit('.', 1)[-1]
    else:
      partial_name = lookup_name
  if isinstance(item, Function):
    return AliasMethod(item, from_constant=isinstance(prev_item, Constant))
  else:
    return item
